from gmat.remma.remma_add import remma_add
from gmat.remma.remma_dom import remma_dom

from gmat.remma.remma_add import _remma_add
from gmat.remma.remma_dom import _remma_dom

from gmat.remma.annotation import annotation_snp_pos
from gmat.remma.random_pair import random_pair, random_pairAD



import numpy as np
from functools import reduce
import gc
import logging
from gmat.uvlmm.design_matrix import design_matrix_wemai_multi_gmat, design_matrix_wemai_multi_gmat_pred


def _wemai_multi_gmat(y, xmat, zmat, gmat_lst, init=None, maxiter=200, cc_par=1.0e-8, cc_gra=1.0e-6):
    """
    Estimate variances for univariate linear mixed model. Multiple genomic relationship matrixes can be included.
    Weighted EM an AI algorithm are used.
    :param y: phenotypic vectors
    :param xmat: designed matrix for fixed effect.
    :param zmat: designed matrix for random effets，csr sparse matrix. y, xmat and zmat can be prepared with
    design_matrix_wemai_multi_gmat program.
    :param gmat_lst: a list of genomic relationship matrixes.
    :param init: initial values for variances. default value is None.
    :param maxiter: the maximal number of interactions. default value is 200.
    :param cc_par: The convergence criteria for update vector.
    :param cc_gra: The convergence criteria for gradient vector
    :return: the estimated variances
    """
    logging.info("#####Prepare#####")
    var_com = [1.0]*(len(gmat_lst)+1)
    if init is not None:
        var_com = init[:]
    var_com = np.array(var_com)
    var_com_new = var_com[:]
    y = np.array(y).reshape(-1, 1)
    n = y.shape[0]
    xmat = np.array(xmat).reshape(n, -1)
    zgmat_lst = []
    for val in range(len(gmat_lst)):
        zgmat_lst.append(zmat.dot((zmat.dot(gmat_lst[val])).T))
    logging.info('Initial variances: ' + ' '.join(list(np.array(var_com_new, dtype=str))))
    logging.info("#####Start the iteration#####\n\n")
    iter = 0
    delta = 1000.0
    cc_gra_val = 1000.0
    cc_par_val = 1000.0
    while iter < maxiter:
        iter += 1
        logging.info('###Round: ' + str(iter) + '###')
        vmat = np.diag([var_com[-1]]*n)
        for val in range(len(zgmat_lst)):
            vmat += zgmat_lst[val]*var_com[val]
        ll_v = np.linalg.slogdet(vmat)[1]
        vmat = np.linalg.inv(vmat)
        # 计算P矩阵
        vxmat = np.dot(vmat, xmat)
        xvxmat = np.dot(xmat.T, vxmat)
        ll_xvx = np.linalg.slogdet(xvxmat)[1]
        xvxmat = np.linalg.inv(xvxmat)
        pmat = reduce(np.dot, [vxmat, xvxmat, vxmat.T])
        pmat = vmat - pmat
        pymat = np.dot(pmat, y)
        # -2logL
        ll_ypy = np.sum(np.dot(y.T, pymat))
        ll_val = -2*(ll_v + ll_xvx + ll_ypy)
        del vmat, vxmat, xvxmat
        gc.collect()
        # 计算一阶偏导、工具变量
        fd_mat = []
        wv_mat = []
        for val in range(len(zgmat_lst)):
            fd_mat_val = -np.trace(np.dot(pmat, zgmat_lst[val])) + reduce(np.dot, [pymat.T, zgmat_lst[val], pymat])
            fd_mat.append(0.5*np.sum(fd_mat_val))
            wv_mat.append(np.dot(zgmat_lst[val], pymat))
        fd_mat_val = -np.trace(pmat) + np.dot(pymat.T, pymat)
        fd_mat.append(0.5 * np.sum(fd_mat_val))
        fd_mat = np.array(fd_mat)
        wv_mat.append(pymat)
        wv_mat = np.concatenate(wv_mat, axis=1)
        # AI矩阵
        ai_mat = 0.5*reduce(np.dot, [wv_mat.T, pmat, wv_mat])
        del pymat, wv_mat, pmat
        gc.collect()
        em_mat = []
        for val in var_com:
            em_mat.append(n/(val*val))
        em_mat = np.diag(em_mat)
        for j in range(0, 101):
            weight = j * 0.01
            wemai_mat = (1 - weight) * ai_mat + weight * em_mat
            delta = np.dot(np.linalg.inv(wemai_mat), fd_mat)
            var_com_new = var_com + delta
            if min(var_com_new) > 0:
                logging.info('EM weight value: ' + str(weight))
                break
        cc_par_val = np.sum(delta * delta) / np.sum(var_com_new * var_com_new)
        cc_par_val = np.sqrt(cc_par_val)
        var_com = var_com_new[:]
        cc_gra_val = np.sqrt(np.sum(fd_mat * fd_mat))
        logging.info('-2logL for last iteration: ' + str(ll_val))
        logging.info('Norm of gradient vector for last iteration: ' + str(cc_gra_val))
        logging.info('Norm of update vector: ' + str(cc_par_val))
        logging.info('Updated variances: ' + ' '.join(list(np.array(var_com_new, dtype=str))))
        if cc_gra_val < cc_gra and cc_par_val < cc_par:
            break
    if cc_gra_val < cc_gra and cc_par_val < cc_par:
        logging.info('Variances converged.')
    else:
        logging.info('Variances not converged.')
    return var_com


def wemai_multi_gmat(pheno_file, bed_file, gmat_lst, init=None, maxiter=200, cc_par=1.0e-8, cc_gra=1.0e-6, out_file='wemai_multi_gmat.var'):
    """
    Estimate variances for univariate linear mixed model. Multiple genomic relationship matrixes can be included.
    Weighted EM an AI algorithm are used.
    :param pheno_file: phenotypic file. The fist two columns are family id, individual id which are same as plink *.fam
    file. The third column is always ones for population mean. The last column is phenotypic values. The ohter covariate
    can be added between columns for population mean and phenotypic values.
    :param bed_file: the prefix for binary file
    :param gmat_lst: a list of genomic relationship matrixes.
    :param init: initial values for variances. default value is None.
    :param maxiter: the maximal number of interactions. default value is 200.
    :param cc_par: The convergence criteria for update vector.
    :param cc_gra: The convergence criteria for gradient vector
    :param out_file: output file to save the estimated variances
    :return: the estimated variances
    """
    y, xmat, zmat = design_matrix_wemai_multi_gmat(pheno_file, bed_file)
    var_com = _wemai_multi_gmat(y, xmat, zmat, gmat_lst, init=init, maxiter=maxiter, cc_par=cc_par, cc_gra=cc_gra)
    np.savetxt(out_file, var_com)
    return var_com


def wemai_multi_gmat_pred(pheno_file, bed_file, gmat_lst, init=None, maxiter=200, cc_par=1.0e-8, cc_gra=1.0e-6, out_file='wemai_multi_gmat_pred'):
    """
    Estimate variances for univariate linear mixed model. Multiple genomic relationship matrixes can be included.
    Weighted EM an AI algorithm are used.
    :param pheno_file: phenotypic file. The fist two columns are family id, individual id which are same as plink *.fam
    file. The third column is always ones for population mean. The last column is phenotypic values. The ohter covariate
    can be added between columns for population mean and phenotypic values.
    :param bed_file: the prefix for binary file
    :param gmat_lst: a list of genomic relationship matrixes.
    :param init: initial values for variances. default value is None.
    :param maxiter: the maximal number of interactions. default value is 200.
    :param cc_par: The convergence criteria for update vector.
    :param cc_gra: The convergence criteria for gradient vector
    :param out_file: output file to save the estimated variances
    :return: the estimated variances
    """
    y, xmat, zmat = design_matrix_wemai_multi_gmat_pred(pheno_file, bed_file)
    var_com = _wemai_multi_gmat(y, xmat, zmat, gmat_lst, init=init, maxiter=maxiter, cc_par=cc_par, cc_gra=cc_gra)
    np.savetxt(out_file + '.var', var_com)
    logging.info('Predict the random effects')
    zgmat_lst = []
    for val in range(len(gmat_lst)):
        zgmat_lst.append(zmat.dot((zmat.dot(gmat_lst[val])).T))
    vmat = np.diag([var_com[-1]] * y.shape[0])
    for val in range(len(zgmat_lst)):
        vmat += zgmat_lst[val] * var_com[val]
    vxmat = np.dot(vmat, xmat)
    xvxmat = np.dot(xmat.T, vxmat)
    xvxmat = np.linalg.inv(xvxmat)
    pmat = reduce(np.dot, [vxmat, xvxmat, vxmat.T])
    pmat = vmat - pmat
    zpymat = zmat.T.dot(np.dot(pmat, y))
    rand_eff = []
    for val in range(len(gmat_lst)):
        eff = np.dot(gmat_lst[val], zpymat) * var_com[val]
        rand_eff.append(eff)
    rand_eff = np.concatenate(rand_eff, axis=1)
    np.savetxt(out_file + '.rand_eff', rand_eff)
    return var_com


def em_mme(y, xmat, gmat_inv, init=None, maxiter=100, cc=1.0e-8):
    var_com = [1.0, 1.0]
    if init is not None:
        var_com = init[:]
    var_com_new = var_com[:]
    # 准备不含方差组分的系数矩阵
    xmat_df, gmat_inv_df = xmat.shape[1], gmat_inv.shape[0]
    y_df = len(y)
    coef_pre = np.identity(xmat_df+gmat_inv_df)
    coef_pre[:xmat_df, :xmat_df] = np.dot(xmat.T, xmat)
    coef_pre[:xmat_df, xmat_df:] = xmat.T
    coef_pre[xmat_df:, :xmat_df] = xmat
    w_mat = np.concatenate((xmat, np.identity(gmat_inv_df)), axis=1)
    # 开始迭代
    iter = 0
    cc_val = 100.0
    while iter < maxiter:
        iter += 1
        print('Round: ', iter)
        # 系数矩阵
        coef = coef_pre/var_com[1]
        coef[xmat_df:, xmat_df:] = coef[xmat_df:, xmat_df:] + gmat_inv/var_com[0]
        # 右手项
        rhs_mat = np.dot(w_mat.T, y)/var_com[1]
        coef_inv = np.linalg.inv(coef)
        eff = np.dot(coef_inv, rhs_mat)
        var_com_new[0] = np.dot(eff[xmat_df:], np.dot(gmat_inv, eff[xmat_df:])) + \
                     np.trace(np.dot(gmat_inv, coef_inv[xmat_df:, xmat_df:]))
        var_com_new[0] = var_com_new[0]/gmat_inv_df
        e_hat = y - np.dot(xmat, eff[:xmat_df]) - eff[xmat_df:]
        var_com_new[1] = np.dot(e_hat, e_hat) + np.trace(np.dot(w_mat, np.dot(coef_inv, w_mat.T)))
        var_com_new[1] = var_com_new[1]/y_df
        print('Updated variances:', var_com_new)
        delta = np.array(var_com_new) - np.array(var_com)
        cc_val = np.sum(delta*delta)/np.sum(np.array(var_com_new)*np.array(var_com_new))
        cc_val = np.sqrt(cc_val)
        var_com = var_com_new[:]
        if cc_val < cc:
            break
    if cc_val < cc:
        print('Variances converged.')
    else:
        print('Variances not converged.')
    return var_com


def pxem_mme(y, xmat, gmat_inv, init=None, maxiter=100, cc=1.0e-8):
    var_com = [1.0, 1.0]
    if init is not None:
        var_com = init[:]
    var_com_new = var_com[:]
    # 准备不含方差组分的系数矩阵
    xmat_df, gmat_inv_df = xmat.shape[1], gmat_inv.shape[0]
    y_df = len(y)
    coef_pre = np.identity(xmat_df+gmat_inv_df)
    coef_pre[:xmat_df, :xmat_df] = np.dot(xmat.T, xmat)
    coef_pre[:xmat_df, xmat_df:] = xmat.T
    coef_pre[xmat_df:, :xmat_df] = xmat
    w_mat = np.concatenate((xmat, np.identity(gmat_inv_df)), axis=1)
    # 开始迭代
    iter = 0
    cc_val = 100.0
    while iter < maxiter:
        iter += 1
        print('Round: ', iter)
        # 系数矩阵
        coef = coef_pre/var_com[1]
        coef[xmat_df:, xmat_df:] = coef[xmat_df:, xmat_df:] + gmat_inv/var_com[0]
        # 右手项
        rhs_mat = np.dot(w_mat.T, y)/var_com[1]
        coef_inv = np.linalg.inv(coef)
        eff = np.dot(coef_inv, rhs_mat)
        var_com_new[0] = np.dot(eff[xmat_df:], np.dot(gmat_inv, eff[xmat_df:])) + \
                     np.trace(np.dot(gmat_inv, coef_inv[xmat_df:, xmat_df:]))
        var_com_new[0] = var_com_new[0]/gmat_inv_df
        e_hat = y - np.dot(xmat, eff[:xmat_df]) - eff[xmat_df:]
        var_com_new[1] = np.dot(e_hat, e_hat) + np.trace(np.dot(w_mat, np.dot(coef_inv, w_mat.T)))
        var_com_new[1] = var_com_new[1]/y_df
        gamma1 = np.dot(eff[xmat_df:], y - np.dot(xmat, eff[:xmat_df])) - np.trace(np.dot(xmat, coef_inv[:xmat_df, xmat_df:]))
        gamma2 = np.dot(eff[xmat_df:], eff[xmat_df:]) + np.trace(coef_inv[xmat_df:, xmat_df:])
        gamma = gamma1/gamma2
        var_com_new[0] = var_com_new[0]*gamma*gamma
        print('Updated variances:', var_com_new)
        delta = np.array(var_com_new) - np.array(var_com)
        cc_val = np.sum(delta*delta)/np.sum(np.array(var_com_new)*np.array(var_com_new))
        cc_val = np.sqrt(cc_val)
        var_com = var_com_new[:]
        if cc_val < cc:
            break
    if cc_val < cc:
        print('Variances converged.')
    else:
        print('Variances not converged.')
    return var_com


def ai_mme(y, xmat, gmat_inv, init=None, maxiter=100, cc=1.0e-8):
    var_com = [1.0, 1.0]
    if init is not None:
        var_com = init[:]
    var_com = np.array(var_com)
    # 准备不含方差组分的系数矩阵
    xmat_df, gmat_inv_df = xmat.shape[1], gmat_inv.shape[0]
    y_df = len(y)
    coef_pre = np.identity(xmat_df + gmat_inv_df)
    coef_pre[:xmat_df, :xmat_df] = np.dot(xmat.T, xmat)
    coef_pre[:xmat_df, xmat_df:] = xmat.T
    coef_pre[xmat_df:, :xmat_df] = xmat
    w_mat = np.concatenate((xmat, np.identity(gmat_inv_df)), axis=1)
    # 开始迭代
    iter = 0
    cc_val = 100.0
    fd_mat = np.zeros(2)
    # ai_mat = np.zeros((2, 2))
    while iter < maxiter:
        iter += 1
        print('Round: ', iter)
        # 系数矩阵
        coef = coef_pre/var_com[1]
        coef[xmat_df:, xmat_df:] = coef[xmat_df:, xmat_df:] + gmat_inv/var_com[0]
        # 右手项
        rhs_mat = np.dot(w_mat.T, y)/var_com[1]
        coef_inv = np.linalg.inv(coef)
        eff = np.dot(coef_inv, rhs_mat)
        e_hat = y - np.dot(xmat, eff[:xmat_df]) - eff[xmat_df:]
        fd_mat[0] = gmat_inv_df/var_com[0] - np.trace(np.dot(coef_inv[xmat_df:, xmat_df:], gmat_inv))/(var_com[0]*var_com[0]) - \
            np.dot(eff[xmat_df:], np.dot(gmat_inv, eff[xmat_df:]))/(var_com[0]*var_com[0])
        fd_mat[1] = y_df/var_com[1] - np.trace(np.dot(np.dot(coef_inv, w_mat.T), w_mat))/(var_com[1]*var_com[1]) - \
            np.dot(e_hat, e_hat)/(var_com[1]*var_com[1])
        fd_mat = -0.5*fd_mat
        h_mat1 = np.array(eff[xmat_df:]/var_com[0]).reshape(gmat_inv_df, 1)
        h_mat2 = np.array(e_hat / var_com[1]).reshape(y_df, 1)
        h_mat = np.concatenate((h_mat1, h_mat2), axis=1)
        qrq = np.divide(np.dot(h_mat.T, h_mat), var_com[-1])
        left = np.divide(np.dot(w_mat.T, h_mat), var_com[-1])
        eff_h_mat = np.dot(coef_inv, left)
        ai_mat = np.subtract(qrq, np.dot(left.T, eff_h_mat))
        ai_mat = 0.5 * ai_mat
        ai_mat_inv = np.linalg.inv(ai_mat)
        var_com_new = var_com + np.dot(ai_mat_inv, fd_mat)
        print('Updated variances:', var_com_new)
        delta = np.array(var_com_new) - np.array(var_com)
        cc_val = np.sum(delta * delta) / np.sum(np.array(var_com_new) * np.array(var_com_new))
        cc_val = np.sqrt(cc_val)
        var_com = var_com_new[:]
        if cc_val < cc:
            break
    if cc_val < cc:
        print('Variances converged.')
    else:
        print('Variances not converged.')
    return var_com


def emai_mme(y, xmat, gmat_inv, init=None, maxiter=100, cc=1.0e-8):
    var_com = [1.0, 1.0]
    if init is not None:
        var_com = init[:]
    var_com = np.array(var_com)
    var_com_new = var_com[:]
    # 准备不含方差组分的系数矩阵
    xmat_df, gmat_inv_df = xmat.shape[1], gmat_inv.shape[0]
    y_df = len(y)
    coef_pre = np.identity(xmat_df + gmat_inv_df)
    coef_pre[:xmat_df, :xmat_df] = np.dot(xmat.T, xmat)
    coef_pre[:xmat_df, xmat_df:] = xmat.T
    coef_pre[xmat_df:, :xmat_df] = xmat
    w_mat = np.concatenate((xmat, np.identity(gmat_inv_df)), axis=1)
    # 开始迭代
    iter = 0
    cc_val = 100.0
    fd_mat = np.zeros(2)
    em_mat = np.zeros((2, 2))
    while iter < maxiter:
        iter += 1
        print('Round: ', iter)
        # 系数矩阵
        coef = coef_pre/var_com[1]
        coef[xmat_df:, xmat_df:] = coef[xmat_df:, xmat_df:] + gmat_inv/var_com[0]
        # 右手项
        rhs_mat = np.dot(w_mat.T, y)/var_com[1]
        coef_inv = np.linalg.inv(coef)
        eff = np.dot(coef_inv, rhs_mat)
        e_hat = y - np.dot(xmat, eff[:xmat_df]) - eff[xmat_df:]
        fd_mat[0] = gmat_inv_df/var_com[0] - np.trace(np.dot(coef_inv[xmat_df:, xmat_df:], gmat_inv))/(var_com[0]*var_com[0]) - \
            np.dot(eff[xmat_df:], np.dot(gmat_inv, eff[xmat_df:]))/(var_com[0]*var_com[0])
        fd_mat[1] = y_df/var_com[1] - np.trace(np.dot(np.dot(coef_inv, w_mat.T), w_mat))/(var_com[1]*var_com[1]) - \
            np.dot(e_hat, e_hat)/(var_com[1]*var_com[1])
        fd_mat = -0.5*fd_mat
        h_mat1 = np.array(eff[xmat_df:]/var_com[0]).reshape(gmat_inv_df, 1)
        h_mat2 = np.array(e_hat / var_com[1]).reshape(y_df, 1)
        h_mat = np.concatenate((h_mat1, h_mat2), axis=1)
        qrq = np.divide(np.dot(h_mat.T, h_mat), var_com[-1])
        left = np.divide(np.dot(w_mat.T, h_mat), var_com[-1])
        eff_h_mat = np.dot(coef_inv, left)
        ai_mat = np.subtract(qrq, np.dot(left.T, eff_h_mat))
        ai_mat = 0.5 * ai_mat
        print(fd_mat, ai_mat)
        em_mat[0, 0] = gmat_inv_df / (var_com[0] * var_com[0])
        em_mat[1, 1] = y_df / (var_com[1] * var_com[1])
        for j in range(0, 51):
            weight = j * 0.1
            wemai_mat = (1 - weight) * ai_mat + weight * em_mat
            delta = np.dot(np.linalg.inv(wemai_mat), fd_mat)
            var_com_new = var_com + delta
            if min(var_com_new) > 0:
                print('EM weight value:', weight)
                break
        print('Updated variances:', var_com_new)
        delta = np.array(var_com_new) - np.array(var_com)
        cc_val = np.sum(delta * delta) / np.sum(np.array(var_com_new) * np.array(var_com_new))
        cc_val = np.sqrt(cc_val)
        var_com = var_com_new[:]
        if cc_val < cc:
            break
    if cc_val < cc:
        print('Variances converged.')
    else:
        print('Variances not converged.')
    return var_com


def pxemai_mme(y, xmat, gmat_inv, init=None, maxiter=100, cc=1.0e-8):
    var_com = [1.0, 1.0]
    if init is not None:
        var_com = init[:]
    var_com = np.array(var_com)
    var_com_new = var_com[:]
    # 准备不含方差组分的系数矩阵
    xmat_df, gmat_inv_df = xmat.shape[1], gmat_inv.shape[0]
    y_df = len(y)
    coef_pre = np.identity(xmat_df + gmat_inv_df)
    coef_pre[:xmat_df, :xmat_df] = np.dot(xmat.T, xmat)
    coef_pre[:xmat_df, xmat_df:] = xmat.T
    coef_pre[xmat_df:, :xmat_df] = xmat
    w_mat = np.concatenate((xmat, np.identity(gmat_inv_df)), axis=1)
    # 开始迭代
    iter = 0
    cc_val = 100.0
    weight = 0.0
    fd_mat = np.zeros(2)
    em_mat = np.zeros((2, 2))
    while iter < maxiter:
        iter += 1
        print('Round: ', iter)
        # 系数矩阵
        coef = coef_pre/var_com[1]
        coef[xmat_df:, xmat_df:] = coef[xmat_df:, xmat_df:] + gmat_inv/var_com[0]
        # 右手项
        rhs_mat = np.dot(w_mat.T, y)/var_com[1]
        coef_inv = np.linalg.inv(coef)
        eff = np.dot(coef_inv, rhs_mat)
        e_hat = y - np.dot(xmat, eff[:xmat_df]) - eff[xmat_df:]
        fd_mat[0] = gmat_inv_df/var_com[0] - np.trace(np.dot(coef_inv[xmat_df:, xmat_df:], gmat_inv))/(var_com[0]*var_com[0]) - \
            np.dot(eff[xmat_df:], np.dot(gmat_inv, eff[xmat_df:]))/(var_com[0]*var_com[0])
        fd_mat[1] = y_df/var_com[1] - np.trace(np.dot(np.dot(coef_inv, w_mat.T), w_mat))/(var_com[1]*var_com[1]) - \
            np.dot(e_hat, e_hat)/(var_com[1]*var_com[1])
        fd_mat = -0.5*fd_mat
        h_mat1 = np.array(eff[xmat_df:]/var_com[0]).reshape(gmat_inv_df, 1)
        h_mat2 = np.array(e_hat / var_com[1]).reshape(y_df, 1)
        h_mat = np.concatenate((h_mat1, h_mat2), axis=1)
        qrq = np.divide(np.dot(h_mat.T, h_mat), var_com[-1])
        left = np.divide(np.dot(w_mat.T, h_mat), var_com[-1])
        eff_h_mat = np.dot(coef_inv, left)
        ai_mat = np.subtract(qrq, np.dot(left.T, eff_h_mat))
        ai_mat = 0.5 * ai_mat
        print(fd_mat, ai_mat)
        em_mat[0, 0] = gmat_inv_df / (var_com[0] * var_com[0])
        em_mat[1, 1] = y_df / (var_com[1] * var_com[1])
        for j in range(0, 51):
            weight = j * 0.1
            wemai_mat = (1 - weight) * ai_mat + weight * em_mat
            delta = np.dot(np.linalg.inv(wemai_mat), fd_mat)
            var_com_new = var_com + delta
            if min(var_com_new) > 0:
                print('EM weight value:', weight)
                break
        if weight > 0.001:
            gamma1 = np.dot(eff[xmat_df:], y - np.dot(xmat, eff[:xmat_df])) - np.trace(
                np.dot(xmat, coef_inv[:xmat_df, xmat_df:]))
            gamma2 = np.dot(eff[xmat_df:], eff[xmat_df:]) + np.trace(coef_inv[xmat_df:, xmat_df:])
            gamma = gamma1 / gamma2
            var_com_new[0] = var_com_new[0] * gamma * gamma
        print('Updated variances:', var_com_new)
        delta = np.array(var_com_new) - np.array(var_com)
        cc_val = np.sum(delta * delta) / np.sum(np.array(var_com_new) * np.array(var_com_new))
        cc_val = np.sqrt(cc_val)
        var_com = var_com_new[:]
        if cc_val < cc:
            break
    if cc_val < cc:
        print('Variances converged.')
    else:
        print('Variances not converged.')
    return var_com





import numpy as np
from scipy import linalg
from functools import reduce
import logging
from scipy.sparse import hstack


def pxem_mme(y, xmat, zmat_lst, gmat_inv_lst, init=None, maxiter=100, cc_par=1.0e-8):
    """
    Estimate variance parameters with mme based on pxem
    :param y: Phenotypic vector
    :param xmat: the design matrix for fixed effect
    :param zmat_lst: A list of design matrix for random effects
    :param gmat_inv_lst: A list for the inversion of relationship matrix
    :param init: Default is None. A list for initial values of variance components
    :param maxiter: Default is 100. The maximum number of iteration times.
    :param cc_par: Convergence criteria for update vector.
    :return: The estimated variances.
    """
    num_var = len(gmat_inv_lst) + 1
    var_com = np.ones(num_var)
    if init is not None:
        var_com = np.array(init)
    var_com_new = np.array(var_com) * 100
    y = np.array(y).reshape(-1, 1)
    n = y.shape[0]
    xmat = np.array(xmat).reshape(n, -1)
    xmat_df = xmat.shape[1]
    df_index = [xmat_df]  # the index for the degree of freedom for different effects.
    for i in range(num_var - 1):
        df_index.append(df_index[-1] + gmat_inv_lst[i].shape[0])
    logging.info("Prepare the coefficient matrix without variance parameters")
    zmat_concat = hstack(zmat_lst)
    xzmat = hstack([xmat, zmat_concat])
    coef_null = xzmat.T.dot(xzmat)  # the coefficient matrix
    rhs_null = xzmat.T.dot(y)  # the right hand matrix
    logging.info('Initial variances: ' + ' '.join(list(np.array(var_com, dtype=str))))
    logging.info("Start iteration")
    iter = 0
    cc_par_val = 10000.0
    while iter < maxiter:
        iter += 1
        logging.info('Round: {:d}'.format(iter))
        # coef matrix
        coef = coef_null.toarray() / var_com[-1]
        for k in range(num_var-1):
            indexa = df_index[k]
            indexb = df_index[k+1]
            coef[indexa:indexb, indexa:indexb] = coef[indexa:indexb, indexa:indexb] + gmat_inv_lst[k]/var_com[k]
        # right hand
        rhs_mat = rhs_null/var_com[-1]
        coef_inv = linalg.inv(coef)
        eff = np.dot(coef_inv, rhs_mat)
        # error variance
        e_hat = y - xzmat.dot(eff)
        var_com_new[-1] = np.sum(np.dot(e_hat.T, e_hat)) + \
                          np.sum((xzmat.T.dot(xzmat)).multiply(coef_inv))  # fast trace calculation
        var_com_new[-1] /= y.shape[0]
        # random variances
        y_xb = y - np.dot(xmat, eff[:xmat_df, :])
        gamma_mat = np.zeros((num_var - 1, num_var - 1))
        gamma_vec = np.zeros(num_var - 1)
        for k in range(num_var-1):
            indexa = df_index[k]
            indexb = df_index[k + 1]
            var_com_new[k] = np.sum(coef_inv[indexa:indexb, indexa:indexb]*gmat_inv_lst[k]) \
                             + np.sum(reduce(np.dot, [eff[indexa:indexb, :].T, gmat_inv_lst[k], eff[indexa:indexb, :]]))
            qk = gmat_inv_lst[k].shape[0]
            var_com_new[k] /= qk
            zu1 = zmat_lst[k].dot(eff[indexa:indexb, :])
            gamma1 = np.sum(np.dot(zu1.T, y_xb)) - \
                     np.sum(zmat_lst[k].T.dot(xmat) * coef_inv[indexa:indexb, :xmat_df])
            gamma_vec[k] = gamma1/var_com[-1]
            for m in range(k+1):
                indexc = df_index[m]
                indexd = df_index[m + 1]
                zu2 = zmat_lst[m].dot(eff[indexc:indexd, :])
                gamma2 = np.sum(np.dot(zu1.T, zu2)) + \
                         np.sum((zmat_lst[k].T.dot(zmat_lst[m])).multiply(coef_inv[indexa:indexb, indexc:indexd]))
                gamma_mat[k, m] = gamma2/var_com[-1]
        gamma_mat = gamma_mat + np.tril(gamma_mat, k=-1).T
        # print(gamma_mat, gamma_vec)
        gamma = np.dot(linalg.inv(gamma_mat), gamma_vec)
        var_com_new[:-1] = var_com_new[:-1] * gamma * gamma
        # cc
        delta = np.array(var_com_new) - np.array(var_com)
        cc_par_val = np.sum(delta*delta)/np.sum(np.array(var_com_new)*np.array(var_com_new))
        cc_par_val = np.sqrt(cc_par_val)
        var_com = np.array(var_com_new)
        logging.info('Norm of update vector: {:e}'.format(cc_par_val))
        var_com_str = ' '.join(np.array(var_com, dtype=str))
        logging.info('Updated variances: {}'.format(var_com_str))
        if cc_par_val < cc_par:
            break
    if cc_par_val < cc_par:
        logging.info('Variances converged.')
    else:
        logging.info('Variances not converged.')
    return var_com


def pxem_vmat(y, xmat, zmat_lst, gmat_lst, init=None, maxiter=100, cc_par=1.0e-8):
    """
    Estimate variance parameters with vmat based on pxem
    :param y: Phenotypic vector
    :param xmat: the design matrix for fixed effect
    :param zmat_lst: A list of design matrix for random effects
    :param gmat_lst: A list for relationship matrix
    :param init: Default is None. A list for initial values of variance components
    :param maxiter: Default is 100. The maximum number of iteration times.
    :param cc_par: Convergence criteria for update vector.
    :return: The estimated variances.
    """
    num_var = len(gmat_lst) + 1
    var_com = np.ones(num_var)
    if init is not None:
        var_com = np.array(init)
    var_com_new = np.array(var_com) * 100
    y = np.array(y).reshape(-1, 1)
    n = y.shape[0]
    xmat = np.array(xmat).reshape(n, -1)
    zgzmat_lst = []
    for val in range(num_var - 1):
        zgzmat_lst.append(zmat_lst[val].dot((zmat_lst[val].dot(gmat_lst[val])).T))
    logging.info('Initial variances: ' + ' '.join(list(np.array(var_com, dtype=str))))
    logging.info("#####Start the iteration#####\n\n")
    iter = 0
    cc_par_val = 1000.0
    while iter < maxiter:
        iter += 1
        logging.info('###Round: ' + str(iter) + '###')
        # V, the inverse of V and P
        vmat = np.diag([var_com[-1]] * n)
        for val in range(len(zgzmat_lst)):
            vmat += zgzmat_lst[val] * var_com[val]
        vmat = linalg.inv(vmat)
        vxmat = np.dot(vmat, xmat)
        xvxmat = np.dot(xmat.T, vxmat)
        xvxmat = linalg.inv(xvxmat)
        pmat = vmat - reduce(np.dot, [vxmat, xvxmat, vxmat.T])
        pymat = np.dot(pmat, y)
        # Updated variances
        var_com_new[-1] = var_com[-1] - var_com[-1] * var_com[-1] * np.trace(pmat)/n + \
                          var_com[-1] * var_com[-1] * np.sum(np.dot(pymat.T, pymat))/n
        # random
        y_xb = y - reduce(np.dot, [xmat, xvxmat, vxmat.T, y])
        gamma_mat = np.zeros((num_var - 1, num_var - 1))
        gamma_vec = np.zeros(num_var - 1)
        for k in range(num_var - 1):
            var_com_new[k] = -np.sum(zgzmat_lst[k] * pmat) + np.sum(reduce(np.dot, [pymat.T, zgzmat_lst[k], pymat]))
            var_com_new[k] = var_com[k] + var_com[k] * var_com[k]/gmat_lst[k].shape[0] * var_com_new[k]
            zu1 = np.dot(zgzmat_lst[k], pymat) * var_com[k]
            gamma1 = np.sum(np.dot(zu1.T, y_xb)) + \
                     np.sum(zmat_lst[k].T.dot(xmat) * reduce(np.dot, [(zmat_lst[k].dot(gmat_lst[k])).T, vxmat, xvxmat])) * var_com[k]
            gamma_vec[k] = gamma1/var_com[-1]
            for m in range(k+1):
                zu2 = np.dot(zgzmat_lst[m], pymat) * var_com[m]
                zjgizi = zmat_lst[m].dot((zmat_lst[k].dot(gmat_lst[k])).T) * var_com[k]
                zjgjzi = zmat_lst[m].dot((zmat_lst[k].dot(gmat_lst[m])).T) * var_com[m]
                if k == m:
                    gamma2_trace = np.trace(zjgizi) - np.sum(zjgizi * np.dot(zjgjzi.T, pmat))
                else:
                    gamma2_trace = -np.sum(zjgizi * np.dot(zjgjzi.T, pmat))
                gamma2 = np.sum(np.dot(zu1.T, zu2)) + gamma2_trace
                gamma_mat[k, m] = gamma2 / var_com[-1]
        gamma_mat = gamma_mat + np.tril(gamma_mat, k=-1).T
        # print(gamma_mat, gamma_vec)
        gamma = np.dot(linalg.inv(gamma_mat), gamma_vec)
        var_com_new[:-1] = var_com_new[:-1] * gamma * gamma
        delta = var_com_new - var_com
        cc_par_val = np.sum(delta * delta) / np.sum(var_com_new * var_com_new)
        cc_par_val = np.sqrt(cc_par_val)
        var_com = np.array(var_com_new)
        logging.info('Norm of update vector: ' + str(cc_par_val))
        logging.info('Updated variances: ' + ' '.join(list(np.array(var_com_new, dtype=str))))
        if cc_par_val < cc_par:
            break
    if cc_par_val < cc_par:
        logging.info('Variances converged.')
    else:
        logging.info('Variances not converged.')
    return var_com




import logging
import gc
import numpy as np
import pandas as pd
from gmat.process_plink.process_plink import read_plink, impute_geno


def simu_epistasis_freq(bed_file, add_file, dom_file, epiAA_file, epiAD_file, epiDD_file, ratio=None, mean=1.0, res_var=1.0, out_file='simu_epistasis_freq'):
    logging.info("Read the SNP")
    snp_mat = read_plink(bed_file)
    num_id, num_snp = snp_mat.shape
    if np.any(np.isnan(snp_mat)):
        logging.warning('Missing genotypes are imputed with random genotypes.')
        snp_mat = impute_geno(snp_mat)
    freq = np.sum(snp_mat, axis=0) / (2 * num_id)
    freq.shape = (1, num_snp)
    snp_matA = snp_mat - 2 * freq
    snp_mat[snp_mat > 1.5] = 0.0  # 2替换为0, 变为0、1、0编码
    snp_matD = snp_mat - 2 * freq * (1 - freq)
    del snp_mat
    gc.collect()
    if ratio is None:
        ratio = np.array([2.0, 1.0, 0.5, 0.5, 0.5, 1.0])
    else:
        ratio = np.array(ratio)
    logging.info("Additive")
    add_snp = pd.read_csv(add_file, header=None, sep='\s+')
    add_var = ratio[0]/ratio[-1]*res_var
    freq_a = freq[0, add_snp.iloc[:, 0]]
    add_snp_eff_var = 2*freq_a*(1-freq_a)*np.array(add_snp.iloc[:, 1])*np.array(add_snp.iloc[:, 1])
    add_snp.iloc[:, 1] = add_snp.iloc[:, 1]/np.sqrt(np.sum(add_snp_eff_var)/add_var)
    add_snp.to_csv(add_file + '.norm', sep=' ', header=False, index=False)
    logging.info("Dominance")
    dom_snp = pd.read_csv(dom_file, header=None, sep='\s+')
    dom_var = ratio[1] / ratio[-1] * res_var
    freq_d = freq[0, dom_snp.iloc[:, 0]]
    dom_snp_eff_var = 2 * freq_d * (1 - freq_d) * (1 - 2 * freq_d * (1 - freq_d)) * np.array(dom_snp.iloc[:, 1]) * np.array(dom_snp.iloc[:, 1])
    dom_snp.iloc[:, 1] = dom_snp.iloc[:, 1] / np.sqrt(np.sum(dom_snp_eff_var)/dom_var)
    dom_snp.to_csv(dom_file + '.norm', sep=' ', header=False, index=False)
    logging.info("Additive by additive epistasis")
    epiAA_snp = pd.read_csv(epiAA_file, header=None, sep='\s+')
    epiAA_var = ratio[2] / ratio[-1] * res_var
    epiAA_snp_eff_var = np.var(snp_matA[:, epiAA_snp.iloc[:, 0]] * snp_matA[:, epiAA_snp.iloc[:, 1]] * np.array(epiAA_snp.iloc[:, 2]), axis=0)
    epiAA_snp.iloc[:, 2] = epiAA_snp.iloc[:, 2]/np.sqrt(np.sum(epiAA_snp_eff_var)/epiAA_var)
    epiAA_snp.to_csv(epiAA_file + '.norm', sep=' ', header=False, index=False)
    logging.info("Additive by dominance epistasis")
    epiAD_snp = pd.read_csv(epiAD_file, header=None, sep='\s+')
    epiAD_var = ratio[3] / ratio[-1] * res_var
    epiAD_snp_eff_var = np.var(
        snp_matA[:, epiAD_snp.iloc[:, 0]] * snp_matD[:, epiAD_snp.iloc[:, 1]] * np.array(epiAD_snp.iloc[:, 2]), axis=0)
    epiAD_snp.iloc[:, 2] = epiAD_snp.iloc[:, 2] / np.sqrt(np.sum(epiAD_snp_eff_var)/epiAD_var)
    epiAD_snp.to_csv(epiAD_file + '.norm', sep=' ', header=False, index=False)
    logging.info("Dominance by dominance epistasis")
    epiDD_snp = pd.read_csv(epiDD_file, header=None, sep='\s+')
    epiDD_var = ratio[3] / ratio[-1] * res_var
    epiDD_snp_eff_var = np.var(
        snp_matD[:, epiDD_snp.iloc[:, 0]] * snp_matD[:, epiDD_snp.iloc[:, 1]] * np.array(epiDD_snp.iloc[:, 2]), axis=0)
    epiDD_snp.iloc[:, 2] = epiDD_snp.iloc[:, 2] / np.sqrt(np.sum(epiDD_snp_eff_var)/epiDD_var)
    epiDD_snp.to_csv(epiDD_file + '.norm', sep=' ', header=False, index=False)
    logging.info("Residual")
    res_vec = np.random.normal(0, np.sqrt(res_var), snp_matA.shape[0])
    np.savetxt(out_file + '.res', res_vec)
    logging.info("Phenotypic values")
    pheno_vec = mean + np.sum(snp_matA[:, add_snp.iloc[:, 0]] * np.array(add_snp.iloc[:, 1]), axis=1) + \
                np.sum(snp_matD[:, dom_snp.iloc[:, 0]] * np.array(dom_snp.iloc[:, 1]), axis=1) + \
    np.sum(snp_matA[:, epiAA_snp.iloc[:, 0]] * snp_matA[:, epiAA_snp.iloc[:, 1]] * np.array(epiAA_snp.iloc[:, 2]), axis=1) + \
    np.sum(snp_matA[:, epiAD_snp.iloc[:, 0]] * snp_matD[:, epiAD_snp.iloc[:, 1]] * np.array(epiAD_snp.iloc[:, 2]), axis=1) + \
    np.sum(snp_matD[:, epiDD_snp.iloc[:, 0]] * snp_matD[:, epiDD_snp.iloc[:, 1]] * np.array(epiDD_snp.iloc[:, 2]), axis=1) + \
    res_vec
    fam_df = pd.read_csv(bed_file + '.fam', sep='\s+', header=None)
    fam_df.iloc[:, 2] = 1
    fam_df.iloc[:, 3] = pheno_vec
    res_df = fam_df.iloc[:, 0:4]
    res_df.to_csv(out_file + '.pheno', sep=' ', header=False, index=False)
    return res_df


def simu_epistasis(bed_file, add_file, dom_file, epiAA_file, epiAD_file, epiDD_file, ratio=None, mean=1.0, res_var=1.0, out_file='simu_epistasis'):
    logging.info("Read the SNP")
    snp_mat = read_plink(bed_file)
    num_id, num_snp = snp_mat.shape
    if np.any(np.isnan(snp_mat)):
        logging.warning('Missing genotypes are imputed with random genotypes.')
        snp_mat = impute_geno(snp_mat)
    freq = np.sum(snp_mat, axis=0) / (2 * num_id)
    freq.shape = (1, num_snp)
    snp_matA = snp_mat - 2 * freq
    snp_mat[snp_mat > 1.5] = 0.0  # 2替换为0, 变为0、1、0编码
    snp_matD = snp_mat - 2 * freq * (1 - freq)
    del snp_mat
    gc.collect()
    if ratio is None:
        ratio = np.array([2.0, 1.0, 0.5, 0.5, 0.5, 1.0])
    else:
        ratio = np.array(ratio)
    logging.info("Additive")
    add_snp = pd.read_csv(add_file, header=None, sep='\s+')
    add_var = ratio[0]/ratio[-1]*res_var
    add_snp_eff_var = np.var(snp_matA[:, add_snp.iloc[:, 0]] * np.array(add_snp.iloc[:, 1]), axis=0)
    add_snp.iloc[:, 1] = add_snp.iloc[:, 1]/np.sqrt(np.sum(add_snp_eff_var)/add_var)
    add_snp.to_csv(add_file + '.norm', sep=' ', header=False, index=False)
    logging.info("Dominance")
    dom_snp = pd.read_csv(dom_file, header=None, sep='\s+')
    dom_var = ratio[1] / ratio[-1] * res_var
    dom_snp_eff_var = np.var(snp_matD[:, dom_snp.iloc[:, 0]] * np.array(dom_snp.iloc[:, 1]), axis=0)
    dom_snp.iloc[:, 1] = dom_snp.iloc[:, 1] / np.sqrt(np.sum(dom_snp_eff_var)/dom_var)
    dom_snp.to_csv(dom_file + '.norm', sep=' ', header=False, index=False)
    logging.info("Additive by additive epistasis")
    epiAA_snp = pd.read_csv(epiAA_file, header=None, sep='\s+')
    epiAA_var = ratio[2] / ratio[-1] * res_var
    epiAA_snp_eff_var = np.var(snp_matA[:, epiAA_snp.iloc[:, 0]] * snp_matA[:, epiAA_snp.iloc[:, 1]] * np.array(epiAA_snp.iloc[:, 2]), axis=0)
    epiAA_snp.iloc[:, 2] = epiAA_snp.iloc[:, 2]/np.sqrt(np.sum(epiAA_snp_eff_var)/epiAA_var)
    epiAA_snp.to_csv(epiAA_file + '.norm', sep=' ', header=False, index=False)
    logging.info("Additive by dominance epistasis")
    epiAD_snp = pd.read_csv(epiAD_file, header=None, sep='\s+')
    epiAD_var = ratio[3] / ratio[-1] * res_var
    epiAD_snp_eff_var = np.var(
        snp_matA[:, epiAD_snp.iloc[:, 0]] * snp_matD[:, epiAD_snp.iloc[:, 1]] * np.array(epiAD_snp.iloc[:, 2]), axis=0)
    epiAD_snp.iloc[:, 2] = epiAD_snp.iloc[:, 2] / np.sqrt(np.sum(epiAD_snp_eff_var)/epiAD_var)
    epiAD_snp.to_csv(epiAD_file + '.norm', sep=' ', header=False, index=False)
    logging.info("Dominance by dominance epistasis")
    epiDD_snp = pd.read_csv(epiDD_file, header=None, sep='\s+')
    epiDD_var = ratio[3] / ratio[-1] * res_var
    epiDD_snp_eff_var = np.var(
        snp_matD[:, epiDD_snp.iloc[:, 0]] * snp_matD[:, epiDD_snp.iloc[:, 1]] * np.array(epiDD_snp.iloc[:, 2]), axis=0)
    epiDD_snp.iloc[:, 2] = epiDD_snp.iloc[:, 2] / np.sqrt(np.sum(epiDD_snp_eff_var)/epiDD_var)
    epiDD_snp.to_csv(epiDD_file + '.norm', sep=' ', header=False, index=False)
    logging.info("Residual")
    res_vec = np.random.normal(0, np.sqrt(res_var), snp_matA.shape[0])
    np.savetxt(out_file + '.res', res_vec)
    logging.info("Phenotypic values")
    pheno_vec = mean + np.sum(snp_matA[:, add_snp.iloc[:, 0]] * np.array(add_snp.iloc[:, 1]), axis=1) + \
                np.sum(snp_matD[:, dom_snp.iloc[:, 0]] * np.array(dom_snp.iloc[:, 1]), axis=1) + \
    np.sum(snp_matA[:, epiAA_snp.iloc[:, 0]] * snp_matA[:, epiAA_snp.iloc[:, 1]] * np.array(epiAA_snp.iloc[:, 2]), axis=1) + \
    np.sum(snp_matA[:, epiAD_snp.iloc[:, 0]] * snp_matD[:, epiAD_snp.iloc[:, 1]] * np.array(epiAD_snp.iloc[:, 2]), axis=1) + \
    np.sum(snp_matD[:, epiDD_snp.iloc[:, 0]] * snp_matD[:, epiDD_snp.iloc[:, 1]] * np.array(epiDD_snp.iloc[:, 2]), axis=1) + \
    res_vec
    fam_df = pd.read_csv(bed_file + '.fam', sep='\s+', header=None)
    fam_df.iloc[:, 2] = 1
    fam_df.iloc[:, 3] = pheno_vec
    res_df = fam_df.iloc[:, 0:4]
    res_df.to_csv(out_file + '.pheno', sep=' ', header=False, index=False)
    return res_df



import numpy as np
from scipy import linalg
from functools import reduce
import logging
from scipy.sparse import hstack


def weight_pxemai_vmat(y, xmat, zmat_lst, gmat_lst, init=None, maxiter=100, cc_par=1.0e-8, step=0.01):
    """
    Estimate variance parameters with vmat based on weighted pxem and ai
    :param y: Phenotypic vector
    :param xmat: the design matrix for fixed effect
    :param zmat_lst: A list of design matrix for random effects
    :param gmat_lst: A list for relationship matrix
    :param init: Default is None. A list for initial values of variance components
    :param maxiter: Default is 100. The maximum number of iteration times.
    :param cc_par: Convergence criteria for update vector.
    :param step: Default is 0.01. The increased step for the weight of pxem and ai
    :return: The estimated variances.
    """
    logging.info("#####Prepare#####")
    num_var = len(gmat_lst) + 1
    var_com = np.ones(num_var)
    if init is not None:
        var_com = np.array(init)
    var_com_new = var_com * 1000
    delta = var_com_new - var_com
    y = np.array(y).reshape(-1, 1)
    n = y.shape[0]
    xmat = np.array(xmat).reshape(n, -1)
    zgzmat_lst = []
    for val in range(len(gmat_lst)):
        zgzmat_lst.append(zmat_lst[val].dot((zmat_lst[val].dot(gmat_lst[val])).T))
    logging.info('Initial variances: ' + ' '.join(list(np.array(var_com, dtype=str))))
    logging.info("#####Start the iteration#####\n\n")
    weight_vec = list(np.arange(0, 1, step)) + [1.0]
    iter = 0
    cc_par_val = 1000.0
    while iter < maxiter:
        iter += 1
        logging.info('Round: {:d}'.format(iter))
        # V, the inverse of V and P
        vmat = np.diag([var_com[-1]] * n)
        for val in range(len(zgzmat_lst)):
            vmat += zgzmat_lst[val] * var_com[val]
        vmat = np.linalg.inv(vmat)
        vxmat = np.dot(vmat, xmat)
        xvxmat = np.dot(xmat.T, vxmat)
        xvxmat = linalg.inv(xvxmat)
        pmat = vmat - reduce(np.dot, [vxmat, xvxmat, vxmat.T])
        pymat = np.dot(pmat, y)
        # first partial derivatives and working variates
        fd_mat = np.zeros(num_var)
        wv_mat = []
        fd_mat[-1] = np.trace(pmat) - np.sum(np.dot(pymat.T, pymat))
        fd_mat[-1] = -0.5 * fd_mat[-1]
        for i in range(num_var - 1):
            fd_mat[i] = np.sum(pmat*zgzmat_lst[i]) - np.sum(reduce(np.dot, [pymat.T, zgzmat_lst[i], pymat]))
            fd_mat[i] = -0.5 * fd_mat[i]
            wv_mat.append(np.dot(zgzmat_lst[i], pymat))
        wv_mat.append(pymat)
        wv_mat = np.concatenate(wv_mat, axis=1)
        ai_mat = 0.5 * reduce(np.dot, [wv_mat.T, pmat, wv_mat])
        # em matrix and weighted matrix
        em_mat = []
        for k in range(num_var - 1):
            em_mat.append(gmat_lst[k].shape[0] / (2 * var_com[k] * var_com[k]))
        em_mat.append(n / (2 * var_com[-1] * var_com[-1]))
        em_mat = np.diag(em_mat)
        break_weight = 0.0
        for weight in weight_vec:
            wemai_mat = weight * em_mat + (1 - weight) * ai_mat
            delta = np.dot(linalg.inv(wemai_mat), fd_mat)
            var_com_new = var_com + delta
            if min(var_com_new) > 0:
                logging.info('EM weight value: ' + str(weight))
                break_weight = weight
                break
        var_com_new = var_com + delta
        # pxem gamma
        if break_weight > 0.0:
            y_xb = y - reduce(np.dot, [xmat, xvxmat, vxmat.T, y])
            gamma_mat = np.zeros((num_var - 1, num_var - 1))
            gamma_vec = np.zeros(num_var - 1)
            for k in range(num_var - 1):
                zu1 = np.dot(zgzmat_lst[k], pymat) * var_com[k]
                gamma1 = np.sum(np.dot(zu1.T, y_xb)) + \
                         np.sum(zmat_lst[k].T.dot(xmat) * reduce(np.dot,
                                                                 [(zmat_lst[k].dot(gmat_lst[k])).T, vxmat, xvxmat])) * \
                         var_com[k]
                gamma_vec[k] = gamma1 / var_com[-1]
                for m in range(k + 1):
                    zu2 = np.dot(zgzmat_lst[m], pymat) * var_com[m]
                    zjgizi = zmat_lst[m].dot((zmat_lst[k].dot(gmat_lst[k])).T) * var_com[k]
                    zjgjzi = zmat_lst[m].dot((zmat_lst[k].dot(gmat_lst[m])).T) * var_com[m]
                    if k == m:
                        gamma2_trace = np.trace(zjgizi) - np.sum(zjgizi * np.dot(zjgjzi.T, pmat))
                    else:
                        gamma2_trace = -np.sum(zjgizi * np.dot(zjgjzi.T, pmat))
                    gamma2 = np.sum(np.dot(zu1.T, zu2)) + gamma2_trace
                    gamma_mat[k, m] = gamma_mat[m, k] = gamma2 / var_com[-1]
            gamma = np.dot(linalg.inv(gamma_mat), gamma_vec)
            var_com_new[:-1] = var_com_new[:-1] * gamma * gamma
            delta = var_com_new - var_com
        cc_par_val = np.sum(delta * delta) / np.sum(np.array(var_com_new) * np.array(var_com_new))
        cc_par_val = np.sqrt(cc_par_val)
        var_com = np.array(var_com_new)
        logging.info('Norm of update vector: {:e}'.format(cc_par_val))
        var_com_str = ' '.join(np.array(var_com, dtype=str))
        logging.info('Updated variances: {}'.format(var_com_str))
        if cc_par_val < cc_par:
            break
    if cc_par_val < cc_par:
        logging.info('Variances converged.')
    else:
        logging.info('Variances not converged.')
    return var_com


def weight_pxemai_mme(y, xmat, zmat_lst, gmat_inv_lst, init=None, maxiter=100, cc_par=1.0e-8, step=0.01):
    """
    Estimate variance parameters with mme based on em
    :param y: Phenotypic vector
    :param xmat: the design matrix for fixed effect
    :param zmat_lst: A list of design matrix for random effects
    :param gmat_inv_lst: A list for the inversion of relationship matrix
    :param init: Default is None. A list for initial values of variance components
    :param maxiter: Default is 100. The maximum number of iteration times.
    :param cc_par: Convergence criteria for update vector.
    :param step: Default is 0.01. The increased step for the weight of em and ai
    :return: The estimated variances.
    """
    num_var = len(gmat_inv_lst) + 1
    var_com = np.ones(num_var)
    if init is not None:
        var_com = np.array(init)
    var_com_new = var_com * 1000
    delta = var_com_new - var_com
    y = np.array(y).reshape(-1, 1)
    n = y.shape[0]
    xmat = np.array(xmat).reshape(n, -1)
    xmat_df = xmat.shape[1]
    df_index = [xmat_df]  # the index for the degree of freedom for different effects.
    for i in range(num_var - 1):
        df_index.append(df_index[-1] + gmat_inv_lst[i].shape[0])
    logging.info("Prepare the coefficient matrix without variance parameters")
    zmat_concat = hstack(zmat_lst)
    xzmat = hstack([xmat, zmat_concat])
    coef_null = xzmat.T.dot(xzmat)  # the coefficient matrix
    rhs_null = xzmat.T.dot(y)  # the right hand matrix
    logging.info('Initial variances: ' + ' '.join(list(np.array(var_com, dtype=str))))
    logging.info("Start iteration")
    weight_vec = list(np.arange(0, 1, step)) + [1.0]
    iter = 0
    cc_par_val = 10000.0
    while iter < maxiter:
        iter += 1
        logging.info('Round: {:d}'.format(iter))
        # coef matrix
        coef = coef_null.toarray() / var_com[-1]
        for k in range(num_var - 1):
            indexa = df_index[k]
            indexb = df_index[k + 1]
            coef[indexa:indexb, indexa:indexb] = coef[indexa:indexb, indexa:indexb] + gmat_inv_lst[k] / var_com[k]
        # right hand
        rhs_mat = rhs_null / var_com[-1]
        coef_inv = linalg.inv(coef)
        eff = np.dot(coef_inv, rhs_mat)
        e_hat = y - xzmat.dot(eff)
        # first partial derivatives and working variates
        fd_mat = np.zeros(num_var)
        wv_mat = []
        for i in range(num_var - 1):
            indexa = df_index[i]
            indexb = df_index[i + 1]
            fd_mat[i] = np.sum(coef_inv[indexa:indexb, indexa:indexb] * gmat_inv_lst[i]) + \
                        np.sum(reduce(np.dot, [eff[indexa:indexb, :].T, gmat_inv_lst[i], eff[indexa:indexb, :]]))
            fd_mat[i] = -0.5 * gmat_inv_lst[i].shape[0]/var_com[i] + 0.5 * fd_mat[i] / (var_com[i] * var_com[i])
            wv_mati = zmat_lst[i].dot(eff[indexa:indexb, :]) / var_com[i]
            wv_mat.append(wv_mati)
        fd_mat[-1] = np.sum((xzmat.T.dot(xzmat)).multiply(coef_inv)) + np.sum(np.dot(e_hat.T, e_hat))
        fd_mat[-1] = -0.5*n/var_com[-1] + 0.5 * fd_mat[-1] / (var_com[-1] * var_com[-1])
        wv_mati = e_hat / var_com[-1]
        wv_mat.append(wv_mati)
        wv_mat = np.concatenate(wv_mat, axis=1)
        wvxz_mat = xzmat.T.dot(wv_mat)/var_com[-1]
        ai_mat = np.dot(wv_mat.T, wv_mat) / var_com[-1] - reduce(np.dot, [wvxz_mat.T, coef_inv, wvxz_mat])
        ai_mat = 0.5 * ai_mat
        em_mat = []
        for k in range(num_var - 1):
            em_mat.append(gmat_inv_lst[k].shape[0] / (2 * var_com[k] * var_com[k]))
        em_mat.append(n / (2 * var_com[-1] * var_com[-1]))
        em_mat = np.diag(em_mat)
        break_weight = 0.0
        for weight in weight_vec:
            wemai_mat = weight * em_mat + (1 - weight) * ai_mat
            delta = np.dot(linalg.inv(wemai_mat), fd_mat)
            var_com_new = var_com + delta
            if min(var_com_new) > 0:
                logging.info('EM weight value: ' + str(weight))
                break_weight = weight
                break
        var_com_new = var_com + delta
        # pxem gamma
        if break_weight > 0.0:
            y_xb = y - np.dot(xmat, eff[:xmat_df, :])
            gamma_mat = np.zeros((num_var - 1, num_var - 1))
            gamma_vec = np.zeros(num_var - 1)
            for k in range(num_var - 1):
                indexa = df_index[k]
                indexb = df_index[k + 1]
                zu1 = zmat_lst[k].dot(eff[indexa:indexb, :])
                gamma1 = np.sum(np.dot(zu1.T, y_xb)) - \
                         np.sum(zmat_lst[k].T.dot(xmat) * coef_inv[indexa:indexb, :xmat_df])
                gamma_vec[k] = gamma1 / var_com[-1]
                for m in range(k + 1):
                    indexc = df_index[m]
                    indexd = df_index[m + 1]
                    zu2 = zmat_lst[m].dot(eff[indexc:indexd, :])
                    gamma2 = np.sum(np.dot(zu1.T, zu2)) + \
                             np.sum((zmat_lst[k].T.dot(zmat_lst[m])).multiply(coef_inv[indexa:indexb, indexc:indexd]))
                    gamma_mat[k, m] = gamma_mat[m, k] = gamma2 / var_com[-1]
            gamma = np.dot(linalg.inv(gamma_mat), gamma_vec)
            var_com_new[:-1] = var_com_new[:-1] * gamma * gamma
            delta = np.array(var_com_new) - np.array(var_com)
        cc_par_val = np.sum(delta * delta) / np.sum(np.array(var_com_new) * np.array(var_com_new))
        cc_par_val = np.sqrt(cc_par_val)
        var_com = np.array(var_com_new)
        logging.info('Norm of update vector: {:e}'.format(cc_par_val))
        var_com_str = ' '.join(np.array(var_com, dtype=str))
        logging.info('Updated variances: {}'.format(var_com_str))
        if cc_par_val < cc_par:
            break
    if cc_par_val < cc_par:
        logging.info('Variances converged.')
    else:
        logging.info('Variances not converged.')
    return var_com



import numpy as np
from collections import defaultdict
def is_int(num):
    try:
        int(num)
        return True
    except ValueError:
        return False

def is_float(num):
    try:
        float(num)
        return True
    except ValueError:
        return False

###declare the 2D dict
def dct_3D():
    return defaultdict(dct_31D)

def dct_31D():
    return defaultdict(dct_32D)

def dct_32D():
    return defaultdict()


###declare the 2D dict
def dct_2D():
    return defaultdict(dct_21D)

def dct_21D():
    return defaultdict()


###declare the 1D dict
def dct_1D():
    return defaultdict()

###triple matrix multiplication a x b x a'

def tri_matT(a, b):
    res = np.dot(a,b)
    res = np.dot(res,a.T)
    return res

###triple matrix multiplication a x b x c

def tri_mat(a, b, c):
    res = np.dot(a, b)
    res = np.dot(res, c)
    return res

###triple matrix multiplication a x D x a'  D is diagonal matrix in row vector form

def Dtri_matT(a, b):
    res = np.multiply(a,b)
    res = np.dot(res,a.T)
    return res

###triple matrix multiplication a x D x c  D is diagonal matrix in row vector form

def Dtri_mat(a, b, c):
    res = np.multiply(a,b)
    res = np.dot(res, c)
    return res

####legendre polynomial
def leg(time, order):
    time = np.array(time, dtype=float).reshape(max(time.shape), 1)
    tmin = min(time)
    tmax = max(time)
    tvec = 2 * (time - tmin) / (tmax - tmin) - 1
    
    pmat = []
    for k in range(order + 1):
        c = int(k / 2)
        j = k
        p = 0
        for r in range(0, c + 1):
            p += np.sqrt((2 * j + 1.0) / 2.0) * pow(0.5, j) * (pow(-1, r) *
                    np.math.factorial(2 * j - 2 * r) / (np.math.factorial(r) *
                    np.math.factorial(j - r) * np.math.factorial(j - 2 * r))) * pow(tvec, j - 2 * r)
        p = np.array(p)
        pmat.append(p)
    return pmat


def leg_mt(time, tmax, tmin, order):
    time = np.array(time, dtype=float).reshape(max(time.shape), 1)
    tvec = 2 * (time - tmin) / (tmax - tmin) - 1
    
    pmat = []
    for k in range(order + 1):
        c = int(k / 2)
        j = k
        p = 0
        for r in range(0, c + 1):
            p += np.sqrt((2 * j + 1.0) / 2.0) * pow(0.5, j) * (pow(-1, r) *
                np.math.factorial(2 * j - 2 * r) / (np.math.factorial(r) *
                np.math.factorial(j - r) * np.math.factorial(j - 2 * r))) * pow(tvec, j - 2 * r)
        p = np.array(p)
        pmat.append(p)
    pmat = np.concatenate(pmat, axis=1)
    return pmat


def cal_xvkvkvx(xmat, vinv, onei, onej, onei2, onej2, leg_tp, kin_dct, var_ind, m, n):
    temp1 = reduce(np.matmul, [onei.T, leg_tp.T, vinv, xmat])
    temp2 = reduce(np.matmul, [onej.T, leg_tp.T, vinv, leg_tp, onei2])
    temp3 = reduce(np.matmul, [onej2.T, leg_tp.T, vinv, xmat])
    xvkvkvx = reduce(np.matmul, [temp1.transpose(0, 2, 1), temp2, temp3])
    xvkvkvx = reduce(np.multiply, [kin_dct[var_ind[m, 0]], kin_dct[var_ind[n, 0]], xvkvkvx])
    xvkvkvx = reduce(np.add, xvkvkvx)
    return xvkvkvx

def cal_xvkvkvx_res(xmat, vinv, onei, onej, leg_tp, kin_dct, var_ind, m):
    temp1 = reduce(np.matmul, [onei.T, leg_tp.T, vinv, xmat])
    temp2 = reduce(np.matmul, [onej.T, leg_tp.T, vinv, vinv, xmat])
    xvkvkvx = reduce(np.matmul, [temp1.transpose(0, 2, 1), temp2])
    xvkvkvx = reduce(np.multiply, [kin_dct[var_ind[m, 0]], xvkvkvx])
    xvkvkvx = reduce(np.add, xvkvkvx)
    return xvkvkvx




def longwas_lm(y, xmat):
    n = y.shape[0]
    r = xmat.shape[1]
    xx = np.linalg.inv(np.dot(xmat.T, xmat))
    yy = np.dot(y.T, y)
    yx = np.dot(y.T, xmat)
    yxxy = reduce(np.dot, [yx, xx, yx.T])
    sigma = np.sum((yy - yxxy)/(n - r))
    eff = np.dot(xx, yx.T)
    eff_var = xx*sigma
    return eff, eff_var, sigma



import numpy as np
from scipy import linalg
import datetime
import pandas as pd
from scipy.stats import chi2
import logging

from .common import *
from .pre_mat import *
from .iter_mat import *


def balance_longwas_emai(y, xmat, leg_tp, kin_eigen_val, init=None, maxiter=20, cc_par=1.0e-8,
                         cc_gra=1.0e-3, em_weight_step=0.02):
    cov_dim = leg_tp.shape[1]
    num_record = y.shape[0] * y.shape[1] * y.shape[2]
    ran_df = len(kin_eigen_val)
    var_com = np.array(init)
    var_com_update = var_com.copy()
    delta = 1000.0
    cov = pre_cov_mat_eigen(cov_dim, var_com)
    if cov is None:
        logging.info('The covariances are not positive defined!')
        exit()
    cov_add, cov_per, cov_res = cov[:]
    
    var_ind = np.tril_indices_from(cov_add)
    a = [0] * len(var_ind[0])
    a.extend([1] * len(var_ind[0]))
    a.append(2)
    b = list(var_ind[0])
    b.extend(b)
    b.append(0)
    c = list(var_ind[1])
    c.extend(c)
    c.append(0)
    var_ind = np.array([a, b, c]).T
    kin_dct = {
        0: kin_eigen_val.reshape(len(kin_eigen_val), 1, 1),
        1: 1.0,
        2: 1.0
    }
    iter_count = 0
    cc_par_val = 10000.0
    cc_gra_val = 10000.0
    while iter_count < maxiter:
        iter_count += 1
        fd_mat, ai_mat = pre_fdai_mat_eigen_glm(y, xmat, leg_tp, kin_eigen_val, cov_add, cov_per, cov_res, var_com,
                                                var_ind, kin_dct)
        em_mat = pre_em_mat_eigen(cov_dim, cov_add, cov_per, ran_df, var_com, num_record)
        gamma = -em_weight_step
        while gamma < 1.0:
            gamma = gamma + em_weight_step
            if gamma >= 1.0:
                gamma = 1.0
            wemai_mat = (1 - gamma) * ai_mat + gamma * em_mat
            delta = np.dot(linalg.inv(wemai_mat), fd_mat)
            var_com_update = var_com + delta
            cov = pre_cov_mat_eigen(cov_dim, var_com_update)
            if cov is not None:
                break
        cov_add, cov_per, cov_res = cov[:]
        # Convergence criteria
        cc_par_val = np.sum(pow(delta, 2)) / np.sum(pow(var_com_update, 2))
        cc_par_val = np.sqrt(cc_par_val)
        cc_gra_val = np.sqrt(np.sum(pow(fd_mat, 2))) / len(var_com)
        var_com = var_com_update.copy()
        if cc_par_val < cc_par and cc_gra_val < cc_gra:
            break
    vinv = np.multiply(kin_eigen_val.reshape(len(kin_eigen_val), 1, 1), tri_matT(leg_tp, cov_add)) + \
           tri_matT(leg_tp, cov_per) + np.diag([np.sum(cov_res)] * leg_tp.shape[0])
    vinv = np.linalg.inv(vinv)
    xvx = reduce(np.matmul, [xmat.transpose(0, 2, 1), vinv, xmat])
    xvx = reduce(np.add, xvx)
    xvx = np.linalg.inv(xvx)
    xvy = np.matmul(xmat.transpose(0, 2, 1), np.matmul(vinv, y))
    xvy = reduce(np.add, xvy)
    b = np.dot(xvx, xvy)
    chi_val = np.sum(np.dot(np.dot(b[-cov_dim:, -1].T, np.linalg.inv(xvx[-cov_dim:, -cov_dim:])), b[-cov_dim:, -1]))
    p_val = chi2.sf(chi_val, cov_dim)
    return cc_par_val, cc_gra_val, b[-cov_dim:, -1], chi_val, p_val, xvx[-cov_dim:, -cov_dim:]



import os
import logging
from scipy import linalg
from scipy.sparse import csr_matrix
import numpy as np
import pandas as pd
import gc

from .common import *
from .balance_emai import *


def balance_varcom(data_file, id, tpoint, trait, kin_file, tfix=None, fix=None, forder=3, rorder=3, na_method='omit',
                   init=None, maxiter=100, cc_par=1.0e-8, cc_gra=1.0e6, em_weight_step=0.001,
                   prefix_outfile='balance_varcom'):
    """
    Estimate variance parameters for balanced longitudinal data.
    :param data_file: the data file. The first row is the variate names whose first initial position is alphabetical.
    For the class variates, the first letter must be capital; for the covariates (continuous variates), the first letter
    must be lowercase.
    :param id: A class variate name which indicates the individual id column in the data file.
    :param tpoint: A list of corresponding time points for phenotypic values.
    :param trait: A list indicating the columns for recorded phenotypic values. The column index starts from 0 in the
    data file.
    :param kin_file: the file for genomic relationship matrix. This file can be produced by
    gmat.gmatrix.agmat function using agmat(bed_file, inv=True, small_val=0.001, out_fmt='id_id_val')
    :param tfix: A class variate name for the time varied fixed effect. Default value is None. The value must be None
    in the current version.
    :param fix: Expression for the time independent fixed effect. Default value is None. The value must be None
    in the current version.
    :param forder: the order of Legendre polynomials for the time varied fixed effect. The default value is 3.
    :param rorder: the order of Legendre polynomials for time varied random effects (additive genetic effects and
    permanent environment effects). The default value is 3.
    :param na_method: The method to deal with missing values. The default value is 'omit'. 'omit' method will delete the
    row with missing values. 'include' method will fill the missing values with the adjacent values.
    :param init: the initial values for the variance parameters. The default value is None.
    :param maxiter: the maximum number of iteration. Default is 100.
    :param cc_par: Convergence criteria for the changed variance parameters. Default is 1.0e-8.
    :param cc_gra: Convergence criteria for the norm of gradient vector. Default is 1.0e6.
    :param em_weight_step: the step of the em weight. Default is 0.001.
    :param prefix_outfile: the prefix for the output file. Default is 'balance_varcom'.
    :return: the estimated variance parameters.
    """
    logging.info('########################################################################')
    logging.info('###Prepare the data for unbalanced longitudinal variances estimation.###')
    logging.info('########################################################################')
    logging.info('***Read the data file***')
    logging.info('Data file: ' + data_file)
    data_df = pd.read_csv(data_file, sep='\s+', header=0)
    logging.info('NA method: ' + na_method)
    if na_method == 'omit':
        data_df = data_df.dropna()
    elif na_method == 'include':
        data_df = data_df.fillna(method='ffill')
        data_df = data_df.fillna(method='bfill')
    else:
        logging.info('na_method does not exist: ' + na_method)
        exit()
    col_names = data_df.columns
    logging.info('The column names of data file: ' + ' '.join(list(col_names)))
    logging.info('Note: Variates beginning with a capital letter is converted into factors.')
    class_vec = []
    for val in col_names:
        if not val[0].isalpha():
            logging.info("The first character of columns names must be alphabet!")
            exit()
        if val[0] == val.capitalize()[0]:
            class_vec.append(val)
            data_df[val] = data_df[val].astype('str')
        else:
            try:
                data_df[val] = data_df[val].astype('float')
            except Exception as e:
                logging.info(e)
                logging.info(val + " column may contain string, please check!")
                exit()
    logging.info('Individual column: ' + id)
    if id not in col_names:
        logging.info(id + ' is not in the data file, please check!')
        exit()
    if id not in class_vec:
        logging.info('The initial letter of {} should be capital'.format(id))
        exit()
    id_in_data_lst = list(data_df[id])
    id_in_data = set(id_in_data_lst)
    if len(id_in_data_lst) != len(id_in_data):
        logging.info('Duplicated ids exit in the data file, please check!')
        exit()
    logging.info('Trait column: ' + ' '.join(np.array(trait, dtype=str)))
    logging.info('Trait column name: ' + ' '.join(list(col_names[trait])))
    if len(set(col_names[trait]) & set(class_vec)) != 0:
        logging.info('Phenotype should not be defined as class variable, please check!')
        exit()
    logging.info('Code factor variables of the data file: ' + ' '.join(list(class_vec)))
    code_dct = {}
    for val in class_vec:
        code_val = 0
        code_dct[val] = {}
        col_vec = []
        for i in range(data_df.shape[0]):
            if data_df[val][i] not in code_dct[val]:
                code_val += 1
                code_dct[val][data_df[val][i]] = str(code_val)
            col_vec.append(code_dct[val][data_df[val][i]])
        data_df[val] = np.array(col_vec)
        data_df[val] = data_df[val].astype('int')
    logging.info('***Read the kinship matrix***')
    with open(kin_file) as fin:
        row = []
        col = []
        kin = []
        id_in_kin = set()
        for line in fin:
            arr = line.split()
            if arr[0] not in code_dct[id] or arr[1] not in code_dct[id]:
                continue
            id_in_kin.add(arr[0])
            id_in_kin.add(arr[1])
            row.append(int(code_dct[id][arr[0]]))
            col.append(int(code_dct[id][arr[1]]))
            kin.append(float(arr[2]))
        kin = csr_matrix((np.array(kin), (np.array(row) - 1, np.array(col) - 1))).toarray()
        kin = np.add(kin, kin.T)
        np.fill_diagonal(kin, 0.5 * np.diag(kin))
        del row, col
        gc.collect()
    logging.info('***Eigen decomposition of kinship matrix***')
    id_not_in_kin = list(id_in_data - id_in_kin)
    if len(id_not_in_kin) != 0:
        logging.info('The ID: {} in the data file is not in the kinship file, please remove these IDs!'.format(' '.join(id_not_in_kin)))
        exit()
    kin_eigen_val, kin_eigen_vec = linalg.eigh(kin)
    logging.info('***Build the design matrix for fixed effect***')
    leg_fix = leg(np.array(tpoint), forder)
    leg_fix = np.array(leg_fix).reshape(forder + 1, 1, len(tpoint))
    leg_fix = np.concatenate([leg_fix] * data_df.shape[0], axis=1)
    xmat_t = leg_fix.copy()
    if tfix is not None:
        logging.info('The parameter tfix should be None in current version.')
        exit()
    xmat_t = np.matmul(np.array([kin_eigen_vec.T]), xmat_t)
    xmat_t = xmat_t.transpose(1, 2, 0)
    if fix is not None:
        logging.info('The parameter fix should be None in current version')
        exit()
    logging.info('***Initial values for variances and rotate the phenotypic values')
    leg_tp = leg(np.array(tpoint), rorder)
    leg_tp = np.concatenate(leg_tp, axis=1)
    y = np.array(data_df.iloc[:, trait])
    cov_dim = leg_tp.shape[1]
    y_var = np.var(y) / (cov_dim*2 + 1)
    var_com = []
    cov_var = np.diag([y_var]*cov_dim)
    var_com.extend(cov_var[np.tril_indices_from(cov_var)])
    var_com.extend(cov_var[np.tril_indices_from(cov_var)])
    var_com.append(y_var)
    if init is None:
        var_com = np.array(var_com)
    else:
        if len(var_com) != len(init):
            logging.info('The length of initial variances should be {}'.format(len(var_com)))
            exit()
        else:
            var_com = np.array(init)
    y = np.dot(kin_eigen_vec.T, y)
    y = y.reshape(data_df.shape[0], len(tpoint), 1)
    logging.info('##########################################################')
    logging.info('###Start the balanced longitudinal variances estimation###')
    logging.info('##########################################################')
    res = balance_emai(y, xmat_t, leg_tp, kin_eigen_val, init=var_com, maxiter=maxiter, cc_par=cc_par,
                       cc_gra=cc_gra, em_weight_step=em_weight_step)
    var_file = prefix_outfile + '.var'
    res.to_csv(var_file, sep=' ', index=False)
    return res



# ext_build.py

from os.path import dirname, join, realpath
import cffi

ffi = cffi.FFI() #生成cffi实例

ffi.cdef("""int read_plink_bed(char *bed_file, long long num_id, long long num_snp, double *marker_mat);""") #函数声明

ffi.set_source('_cread_plink_bed', """
    int read_plink_bed(char *bed_file, long long num_id, long long num_snp, double *marker_mat);
""",
sources=[join("gmat/process_plink", "_read_plink_bed.c")],
)

if __name__ == '__main__':
    ffi.compile(verbose=True)


'''
from os.path import dirname, join, realpath
import cffi

ffi = cffi.FFI() #生成cffi实例

ffi.cdef("""int read_plink_bed(char *bed_file, long long num_id, long long num_snp, double *marker_mat);""") #函数声明

ffi.set_source('_cread_plink_bed', """
    int read_plink_bed(char *bed_file, long long num_id, long long num_snp, double *marker_mat);
""",
sources=[join("gmat/process_plink", "_read_plink_bed.cpp")],
source_extension='.cpp'
)

if __name__ == '__main__':
    ffi.compile(verbose=True)
'''


import numpy as np
import random
from scipy import linalg
from scipy.sparse import csr_matrix
from gmat.process_plink.process_plink import read_plink, impute_geno
import pandas as pd
import datetime
from scipy.stats import chi2
import gc
import logging
from tqdm import tqdm
from functools import reduce

from .common import *


def balance_longwas_trans_permutation(data_file, id, tpoint, trait, kin_file, bed_file, var_com, snp_lst=None,
                                      permutation_lst=None, tfix=None, fix=None,
            forder=3, rorder=3, na_method='omit', prefix_outfile='balance_longwas_trans'):
    """
    Longitudinal GWAS for balanced data.
    :param data_file:the data file. The first row is the variate names whose first initial position is alphabetical.
    For the class variates, the first letter must be capital; for the covariates (continuous variates), the first letter
    must be lowercase.
    :param id: A class variate name which indicates the individual id column in the data file.
    :param tpoint: A list of corresponding time points for phenotypic values.
    :param trait: A list indicating the columns for recorded phenotypic values. The column index starts from 0 in the
    data file.
    :param kin_file: the file for genomic relationship matrix. This file can be produced by
    gmat.gmatrix.agmat function using agmat(bed_file, inv=True, small_val=0.001, out_fmt='id_id_val')
    :param bed_file: the plink binary file
    :param var_com: variances parameters from the balance_varcom function.
    :param snp_lst: A list of snp to test. Default is None.
    :param permutation_lst: the index list for permutation. Default is None ([0, 1000)].
    :param tfix: A class variate name for the time varied fixed effect. Default value is None. The value must be None
    in the current version.
    :param fix: Expression for the time independent fixed effect. Default value is None. The value must be None
    in the current version.
    :param forder: the order of Legendre polynomials for the time varied fixed effect. The default value is 3.
    :param rorder: the order of Legendre polynomials for time varied random effects (additive genetic effects and
    permanent environment effects). The default value is 3.
    :param na_method: The method to deal with missing values. The default value is 'omit'. 'omit' method will delete the
    row with missing values. 'include' method will fill the missing values with the adjacent values.
    :param prefix_outfile: the prefix for the output file. Default is 'balance_longwas_trans'.
    :return: A pandas dataframe of GWAS results.
    """
    logging.info('################################')
    logging.info('###Prepare the related matrix###')
    logging.info('################################')
    logging.info('***Read the data file***')
    logging.info('Data file: ' + data_file)
    data_df = pd.read_csv(data_file, sep='\s+', header=0)
    logging.info('NA method: ' + na_method)
    if na_method == 'omit':
        data_df = data_df.dropna()
    elif na_method == 'include':
        data_df = data_df.fillna(method='ffill')
        data_df = data_df.fillna(method='bfill')
    else:
        logging.info('na_method does not exist: ' + na_method)
        exit()
    col_names = data_df.columns
    logging.info('The column names of data file: ' + ' '.join(list(col_names)))
    logging.info('Note: Variates beginning with a capital letter is converted into factors.')
    class_vec = []
    for val in col_names:
        if not val[0].isalpha():
            logging.info("The first character of columns names must be alphabet!")
            exit()
        if val[0] == val.capitalize()[0]:
            class_vec.append(val)
            data_df[val] = data_df[val].astype('str')
        else:
            try:
                data_df[val] = data_df[val].astype('float')
            except Exception as e:
                logging.info(e)
                logging.info(val + ": may contain string, please check!")
                exit()
    logging.info('Individual column: ' + id)
    if id not in col_names:
        logging.info(id + ' is not in the data file, please check!')
        exit()
    if id not in class_vec:
        logging.info('The initial letter of {} should be capital'.format(id))
        exit()
    id_in_data_lst = list(data_df[id])
    id_in_data = set(id_in_data_lst)
    logging.info('Trait column: ' + ' '.join(np.array(trait, dtype=str)))
    logging.info('Trait column name: ' + ' '.join(list(col_names[trait])))
    if len(set(col_names[trait]) & set(class_vec)) != 0:
        logging.info('Phenotype should not be defined as class variable, please check!')
        exit()
    logging.info('Code factor variables of the data file: ' + ' '.join(list(class_vec)))
    code_dct = {}
    for val in class_vec:
        code_val = 0
        code_dct[val] = {}
        col_vec = []
        for i in range(data_df.shape[0]):
            if data_df[val][i] not in code_dct[val]:
                code_val += 1
                code_dct[val][data_df[val][i]] = str(code_val)
            col_vec.append(code_dct[val][data_df[val][i]])
        data_df[val] = np.array(col_vec)
        data_df[val] = data_df[val].astype('int')
    logging.info('***Read the kinship matrix***')
    with open(kin_file) as fin:
        row = []
        col = []
        kin = []
        id_in_kin = set()
        for line in fin:
            arr = line.split()
            if arr[0] not in code_dct[id] or arr[1] not in code_dct[id]:
                continue
            id_in_kin.add(arr[0])
            id_in_kin.add(arr[1])
            row.append(int(code_dct[id][arr[0]]))
            col.append(int(code_dct[id][arr[1]]))
            kin.append(float(arr[2]))
        kin = csr_matrix((np.array(kin), (np.array(row) - 1, np.array(col) - 1))).toarray()
        kin = np.add(kin, kin.T)
        np.fill_diagonal(kin, 0.5 * np.diag(kin))
        del row, col
        gc.collect()
    logging.info('***Eigen decomposition of kinship matrix***')
    id_not_in_kin = list(id_in_data - id_in_kin)
    if len(id_not_in_kin) != 0:
        logging.info('The ID: {} in the data file is not in the kinship file, please remove these IDs!'.format(' '.join(id_not_in_kin)))
        exit()
    kin_eigen_val, kin_eigen_vec = linalg.eigh(kin)
    logging.info('***Build the design matrix for fixed effect***')
    leg_fix = leg(np.array(tpoint), forder)
    leg_fix = np.array(leg_fix).reshape(forder + 1, 1, len(tpoint))
    leg_fix = np.concatenate([leg_fix] * data_df.shape[0], axis=1)
    xmat_t = leg_fix.copy()
    if tfix is not None:
        logging.info('The parameter tfix should be None in current version.')
        exit()
    xmat_t = np.matmul(np.array([kin_eigen_vec.T]), xmat_t)
    xmat = xmat_t.transpose(1, 2, 0)
    if fix is not None:
        logging.info('The parameter fix should be None in current version')
        exit()
    # T matrix for random effect
    leg_tp = leg(np.array(tpoint), rorder)
    leg_tp = np.concatenate(leg_tp, axis=1)
    y = np.array(data_df.iloc[:, trait])
    y = np.dot(kin_eigen_vec.T, y)
    y = y.reshape(data_df.shape[0], len(tpoint), 1)
    if var_com.shape[0] != rorder*(rorder+1) + 2*(rorder+1) + 1:
        logging.info('Variances do not match the data, please check')
        exit()
    logging.info('***Read the snp data***')
    snp_mat = read_plink(bed_file)
    if np.any(np.isnan(snp_mat)):
        logging.info('Missing genotypes are imputed with random genotypes.')
        snp_mat = impute_geno(snp_mat)
    num_id = snp_mat.shape[0]
    num_snp = snp_mat.shape[1]
    logging.info("There are {:d} individuals and {:d} SNPs.".format(num_id, num_snp))
    fam_df = pd.read_csv(bed_file + '.fam', sep='\s+', header=None)
    id_in_geno = list(np.array(fam_df.iloc[:, 1], dtype=str))
    if len(set(id_in_data_lst) - set(id_in_geno)) != 0:
        logging.info(' '.join(list(set(id_in_data_lst) - set(id_in_geno))) + ' in the data file is not in the snp file!')
        exit()
    # snp list
    if snp_lst is None:
        snp_lst = range(num_snp)
    else:
        try:
            snp_lst = np.array(snp_lst, dtype=int)
        except Exception as e:
            logging.info(e)
            logging.info('The snp list value should be int')
            exit()
    snp_lst = list(snp_lst)
    id_in_data_index = []
    for i in id_in_data_lst:
        id_in_data_index.append(id_in_geno.index(i))
    snp_mat = snp_mat[id_in_data_index, :]
    snp_mat = snp_mat[:, snp_lst]
    freq = np.sum(snp_mat, axis=0) / (2 * num_id)
    freq.shape = (1, len(freq))
    snp_mat = snp_mat - 2 * freq
    snp_mat = np.dot(kin_eigen_vec.T, snp_mat)
    logging.info('***Prepare the variance structure***')
    add_cov = var_com.loc[var_com.loc[:, 'vari']==1, :]
    row = np.array(add_cov['varij']) - 1
    col = np.array(add_cov['varik']) - 1
    val = add_cov['var_val']
    add_cov = csr_matrix((val, (row, col))).toarray()
    add_cov = add_cov + np.tril(add_cov, k=-1).T
    per_cov = var_com.loc[var_com.loc[:, 'vari'] == 2, :]
    row = np.array(per_cov['varij']) - 1
    col = np.array(per_cov['varik']) - 1
    val = per_cov['var_val']
    per_cov = csr_matrix((val, (row, col))).toarray()
    per_cov = per_cov + np.tril(per_cov, k=-1).T
    res_var = np.array(var_com['var_val'])[-1]
    vinv = np.multiply(kin_eigen_val.reshape(len(kin_eigen_val), 1, 1), tri_matT(leg_tp, add_cov)) + \
        tri_matT(leg_tp, per_cov) + np.diag([np.sum(res_var)] * leg_tp.shape[0])
    vinv = np.linalg.inv(vinv)
    vx = np.matmul(vinv, xmat)
    xvx = np.matmul(xmat.transpose(0, 2, 1), vx)
    xvx = reduce(np.add, xvx)
    xvx = np.linalg.inv(xvx)
    xvy = np.matmul(xmat.transpose(0, 2, 1), np.matmul(vinv, y))
    xvy = reduce(np.add, xvy)
    y_xb = y - np.matmul(xmat, np.dot(xvx, xvy))
    py = np.matmul(vinv, y_xb)
    logging.info('########################################################################')
    logging.info('###Start the linear transformation longitudinal GWAS for balance data###')
    logging.info('########################################################################')
    leg_tpoint_mat = leg_mt(np.array(tpoint), max(tpoint), min(tpoint), forder)
    leg_tpoint_accum = np.sum(leg_tpoint_mat, axis=0)
    chi_df = leg_tp.shape[1]
    if permutation_lst is None:
        permutation_lst = range(1000)
    id_perm = list(range(num_id))
    for rep in permutation_lst:
        logging.info("***Permutation: {} ***".format(rep))
        random.shuffle(id_perm)
        snp_mat = snp_mat[id_perm, :]
        eff_vec = []
        chi_vec = []
        p_vec = []
        p_min_vec = []
        p_accum_vec = []
        temp_gt = np.dot(add_cov, leg_tp.T)
        for i in tqdm(range(snp_mat.shape[1])):
            snpi = snp_mat[:, i].reshape(num_id, 1, 1)
            snpi = np.multiply(snpi, temp_gt)
            snp_eff = np.matmul(snpi, py)
            snp_eff = reduce(np.add, snp_eff)
            snp_cov1 = reduce(np.matmul, [snpi, vinv, snpi.transpose(0, 2, 1)])
            snp_cov1 = reduce(np.add, snp_cov1)
            snp_cov2 = np.matmul(snpi, vx)
            snp_cov2 = reduce(np.add, snp_cov2)
            snp_cov2 = reduce(np.dot, [snp_cov2, xvx, snp_cov2.T])
            snp_cov = snp_cov1 - snp_cov2
            chi_val = np.sum(np.dot(np.dot(snp_eff.T, np.linalg.inv(snp_cov)), snp_eff))
            p_val = chi2.sf(chi_val, chi_df)
            eff_vec.append(snp_eff)
            chi_vec.append(chi_val)
            p_vec.append(p_val)
            p_tpoint_vec = []
            for k in range(leg_tpoint_mat.shape[0]):
                eff_tpoint = np.sum(np.dot(leg_tpoint_mat[k, :], snp_eff))
                eff_var_tpoint = np.sum(np.dot(leg_tpoint_mat[k, :], np.dot(snp_cov, leg_tpoint_mat[k, :])))
                chi_tpoint = eff_tpoint * eff_tpoint / eff_var_tpoint
                p_tpoint = chi2.sf(chi_tpoint, 1)
                p_tpoint_vec.append(p_tpoint)
            p_min_vec.append(min(p_tpoint_vec))
            eff_accum = np.sum(np.dot(leg_tpoint_accum, snp_eff))
            eff_var_accum = np.sum(np.dot(leg_tpoint_accum, np.dot(snp_cov, leg_tpoint_accum)))
            chi_accum = eff_accum * eff_accum / eff_var_accum
            p_accum = chi2.sf(chi_accum, 1)
            p_accum_vec.append(p_accum)
        logging.info('Finish association analysis')
        logging.info('***Output***')
        snp_info_file = bed_file + '.bim'
        snp_info = pd.read_csv(snp_info_file, sep='\s+', header=None)
        res_df = snp_info.iloc[snp_lst, [0, 1, 3, 4, 5]]
        res_df.columns = ['chro', 'snp_ID', 'pos', 'allele1', 'allele2']
        res_df.loc[:, 'order'] = snp_lst
        res_df = res_df.iloc[:, [5, 0, 1, 2, 3, 4]]
        eff_vec = np.array(eff_vec)
        for i in range(eff_vec.shape[1]):
            col_ind = 'eff' + str(i)
            res_df.loc[:, col_ind] = eff_vec[:, i]
        res_df.loc[:, 'chi_val'] = chi_vec
        res_df.loc[:, 'p_val'] = p_vec
        res_df.loc[:, 'p_min'] = p_min_vec
        res_df.loc[:, 'p_accum'] = p_accum_vec
        out_file = prefix_outfile + '.' + str(rep)
        res_df.to_csv(out_file, sep=' ', index=False)
    return 0



import numpy as np
from scipy import linalg
import pandas as pd
from functools import reduce
import gc
import time
import logging
import sys
from scipy.stats import chi2

from gmat.process_plink.process_plink import read_plink, impute_geno
from gmat.uvlmm.design_matrix import design_matrix_wemai_multi_gmat


def _remma_dom(y, xmat, zmat, gmat_lst, var_com, bed_file, out_file='remma_dom'):
    """
    Dominance test by random SNP-BLUP model.
    :param y: phenotypic vector
    :param xmat: Designed matrix for fixed effect
    :param zmat: csr sparse matrix. Designed matrix for random effect.
    :param gmat_lst: A list for relationship matrix
    :param var_com: A list of estimated variances. var_com[0]: additive variances; var_com[1]: dominance variances
    :param bed_file: the prefix for plink binary file
    :param out_file: The output file. Default is 'remma_dom'.
    :return: pandas data frame for results.
    """
    logging.info("Calculate the phenotypic covariance matrix and inversion")
    y = np.array(y).reshape(-1, 1)
    n = y.shape[0]
    xmat = np.array(xmat).reshape(n, -1)
    vmat = np.diag([var_com[-1]] * n)
    for val in range(len(gmat_lst)):
        vmat += zmat.dot((zmat.dot(gmat_lst[val])).T) * var_com[val]
    # del gmat_lst
    # gc.collect()
    vmat_inv = linalg.inv(vmat)
    logging.info("Calculate P matrix")
    vxmat = np.dot(vmat_inv, xmat)
    xvxmat = np.dot(xmat.T, vxmat)
    xvxmat = linalg.inv(xvxmat)
    pmat = reduce(np.dot, [vxmat, xvxmat, vxmat.T])
    pmat = vmat_inv - pmat
    pymat = zmat.T.dot(np.dot(pmat, y))
    # pvpmat = reduce(np.dot, [pmat, vmat, pmat])
    pvpmat = zmat.T.dot((zmat.T.dot(pmat)).T)  # pvp = p
    del vmat, vmat_inv, pmat
    gc.collect()
    logging.info("Read the SNP")
    snp_mat = read_plink(bed_file)
    num_id, num_snp = snp_mat.shape
    if np.any(np.isnan(snp_mat)):
        logging.warning('Missing genotypes are imputed with random genotypes.')
        snp_mat = impute_geno(snp_mat)
    freq = np.sum(snp_mat, axis=0) / (2 * num_id)
    freq.shape = (1, num_snp)
    scale_vec = 2 * freq * (1 - freq)
    scale = np.sum(scale_vec * (1 - scale_vec))
    logging.info('The scaled factor is: {:.3f}'.format(scale))
    snp_mat[snp_mat > 1.5] = 0.0
    snp_mat = snp_mat - scale_vec
    eff_vec = np.dot(snp_mat.T, pymat)[:, -1] * var_com[1] / scale
    var_vec = np.sum(snp_mat * np.dot(pvpmat, snp_mat), axis=0) * var_com[1] * var_com[1]/(scale*scale)
    eff_vec_to_fixed = eff_vec * var_com[1] / (var_vec * scale)
    chi_vec = eff_vec*eff_vec/var_vec
    p_vec = chi2.sf(chi_vec, 1)
    snp_info_file = bed_file + '.bim'
    snp_info = pd.read_csv(snp_info_file, sep='\s+', header=None)
    res_df = snp_info.iloc[:, [0, 1, 3, 4, 5]]
    res_df.columns = ['chro', 'snp_ID', 'pos', 'allele1', 'allele2']
    res_df.loc[:, 'eff_val'] = eff_vec
    res_df.loc[:, 'chi_val'] = chi_vec
    res_df.loc[:, 'eff_val_to_fixed'] = eff_vec_to_fixed
    res_df.loc[:, 'p_val'] = p_vec
    try:
        res_df.to_csv(out_file, index=False, header=True, sep=' ')
    except Exception as e:
        logging.error(e)
        sys.exit()
    return res_df


def remma_dom(pheno_file, bed_file, gmat_lst, var_com, out_file='remma_dom'):
    """
    Dominance test by random SNP-BLUP model.
    :param pheno_file: phenotypic file. The fist two columns are family id, individual id which are same as plink *.fam
    file. The third column is always ones for population mean. The last column is phenotypic values. The ohter covariate
    can be added between columns for population mean and phenotypic values.
    :param bed_file: the prefix for binary file
    :param gmat_lst: a list of genomic relationship matrixes.
    :param var_com: Estimated variances
    :param out_file: output file. default value is 'remma_dom'.
    :return: pandas data frame for results.
    """
    y, xmat, zmat = design_matrix_wemai_multi_gmat(pheno_file, bed_file)
    res = _remma_dom(y, xmat, zmat, gmat_lst, var_com, bed_file, out_file=out_file)
    return res



import numpy as np
import pandas as pd
from gmat.gmatrix import agmat
from gmat.longwas.unbalance import unbalance_varcom
import logging

logging.basicConfig(level=logging.INFO)

bed_file = '../data/mouse_long/plink'
kin_lst = agmat(bed_file, inv=True, small_val=0.001, out_fmt='id_id_val')
data_file = '../data/mouse_long/phe.unbalance.txt'
id = 'ID'
tpoint = 'weak'
trait = 'trait'
kin_inv_file = '../data/mouse_long/plink.agiv2'
tfix = 'Sex'
prefix_outfile = '../data/mouse_long/unbalance_varcom'
res_var = unbalance_varcom(data_file, id, tpoint, trait, kin_inv_file, tfix=None, prefix_outfile=prefix_outfile)
print(res_var)


kin_file = '../data/mouse_long/plink.agrm2'
var_com = pd.read_csv("../data/mouse_long/unbalance_varcom.var", sep='\s+', header=0)

from gmat.longwas.unbalance import unbalance_longwas_fixed
prefix_outfile = '../data/mouse_long/unbalance_longwas_fixed'
res_fixed = unbalance_longwas_fixed(data_file, id, tpoint, trait, bed_file, kin_file, var_com,  tfix=None,
                             prefix_outfile=prefix_outfile)

from gmat.longwas.unbalance import unbalance_longwas_fixed_permutation
res_fixed_perm = unbalance_longwas_fixed_permutation(data_file, id, tpoint, trait, bed_file, kin_file, var_com,
                                    tfix=None, prefix_outfile=prefix_outfile)

from gmat.longwas.unbalance import unbalance_longwas_trans
prefix_outfile = '../data/mouse_long/unbalance_longwas_trans'
res_trans = unbalance_longwas_trans(data_file, id, tpoint, trait, bed_file, kin_file, var_com, tfix=None,
                                    prefix_outfile=prefix_outfile)

from gmat.longwas.unbalance import unbalance_longwas_trans_permutation
res_trans_perm = unbalance_longwas_trans_permutation(data_file, id, tpoint, trait, bed_file, kin_file, var_com,
                                    tfix=None, prefix_outfile=prefix_outfile)



import numpy as np
from scipy import linalg
from functools import reduce
import gc
import logging
from scipy.sparse import hstack


def em_mme2(y, xmat, zmat_lst, gmat_inv_lst, init=None, maxiter=100, cc_par=1.0e-8):
    """
    Estimate variance parameters with mme based on em
    :param y: Phenotypic vector
    :param xmat: the design matrix for fixed effect
    :param zmat_lst: A list of design matrix for random effects
    :param gmat_inv_lst: A list for the inversion of relationship matrix
    :param init: Default is None. A list for initial values of variance components
    :param maxiter: Default is 100. The maximum number of iteration times.
    :param cc_par: Convergence criteria for update vector.
    :return: The estimated variances.
    """
    num_var = len(gmat_inv_lst) + 1
    var_com = np.ones(num_var)
    if init is not None:
        var_com = np.array(init)
    var_com_new = np.array(var_com) * 100
    y = np.array(y).reshape(-1, 1)
    n = y.shape[0]
    xmat = np.array(xmat).reshape(n, -1)
    xmat_df = xmat.shape[1]
    df_index = [xmat_df]  # the index for the degree of freedom for different effects.
    for i in range(num_var - 1):
        df_index.append(df_index[-1] + gmat_inv_lst[i].shape[0])
    logging.info("Prepare the coefficient matrix without variance parameters")
    zmat_concat = hstack(zmat_lst)
    xzmat = hstack([xmat, zmat_concat])
    coef_null = xzmat.T.dot(xzmat)  # the coefficient matrix
    rhs_null = xzmat.T.dot(y)  # the right hand matrix
    h_mat = np.eye(n) - reduce(np.dot, [xmat, linalg.inv(np.dot(xmat.T, xmat)), xmat.T])
    whw_mat = xzmat.T.dot((xzmat.T.dot(h_mat)).T)
    logging.info("Start iteration")
    iter = 0
    cc_par_val = 10000.0
    while iter < maxiter:
        iter += 1
        logging.info('Round: {:d}'.format(iter))
        # coef matrix
        coef = coef_null.toarray() / var_com[-1]
        for k in range(num_var - 1):
            indexa = df_index[k]
            indexb = df_index[k + 1]
            coef[indexa:indexb, indexa:indexb] = coef[indexa:indexb, indexa:indexb] + gmat_inv_lst[k] / var_com[k]
        # right hand
        rhs_mat = rhs_null/var_com[-1]
        coef_inv = linalg.inv(coef)
        eff = np.dot(coef_inv, rhs_mat)
        # error variance
        y_zu = y - zmat_concat.dot(eff[xmat_df:, :])
        var_com_new[-1] = np.sum(reduce(np.dot, [y_zu.T, h_mat, y_zu])) + \
                            np.sum(whw_mat * coef_inv)  # fast trace calculation
        var_com_new[-1] /= n - xmat_df
        # random variances
        for k in range(num_var - 1):
            indexa = df_index[k]
            indexb = df_index[k + 1]
            var_com_new[k] = np.sum(coef_inv[indexa:indexb, indexa:indexb]*gmat_inv_lst[k]) \
                             + np.sum(reduce(np.dot, [eff[indexa:indexb, :].T, gmat_inv_lst[k], eff[indexa:indexb, :]]))
            qk = gmat_inv_lst[k].shape[0]
            var_com_new[k] /= qk
        # cc
        delta = np.array(var_com_new) - np.array(var_com)
        cc_par_val = np.sum(delta*delta)/np.sum(np.array(var_com_new)*np.array(var_com_new))
        cc_par_val = np.sqrt(cc_par_val)
        var_com = np.array(var_com_new)
        logging.info('Norm of update vector: {:e}'.format(cc_par_val))
        var_com_str = ' '.join(np.array(var_com, dtype=str))
        logging.info('Updated variances: {}'.format(var_com_str))
        if cc_par_val < cc_par:
            break
    if cc_par_val < cc_par:
        logging.info('Variances converged.')
    else:
        logging.info('Variances not converged.')
    return var_com


def em_vmat2(y, xmat, zmat_lst, gmat_lst, init=None, maxiter=100, cc_par=1.0e-8):
    """
    Estimate variance parameters with vmat based on em
    :param y: Phenotypic vector
    :param xmat: the design matrix for fixed effect
    :param zmat_lst: A list of design matrix for random effects
    :param gmat_lst: A list for relationship matrix
    :param init: Default is None. A list for initial values of variance components
    :param maxiter: Default is 100. The maximum number of iteration times.
    :param cc_par: Convergence criteria for update vector.
    :return: The estimated variances.
    """
    num_var = len(gmat_lst) + 1
    var_com = np.ones(num_var)
    if init is not None:
        var_com = np.array(init)
    var_com_new = np.array(var_com) * 100
    y = np.array(y).reshape(-1, 1)
    n = y.shape[0]
    xmat = np.array(xmat).reshape(n, -1)
    zgzmat_lst = []
    for val in range(num_var - 1):
        zgzmat_lst.append(zmat_lst[val].dot((zmat_lst[val].dot(gmat_lst[val])).T))
    h_mat = np.eye(n) - reduce(np.dot, [xmat, linalg.inv(np.dot(xmat.T, xmat)), xmat.T])
    logging.info('Initial variances: ' + ' '.join(list(np.array(var_com, dtype=str))))
    logging.info("#####Start the iteration#####\n\n")
    iter = 0
    cc_par_val = 1000.0
    while iter < maxiter:
        iter += 1
        logging.info('###Round: ' + str(iter) + '###')
        # V, the inverse of V and P
        vmat = np.diag([var_com[-1]] * n)
        for val in range(len(zgzmat_lst)):
            vmat += zgzmat_lst[val] * var_com[val]
        zgzmat = vmat - np.diag([var_com[-1]] * n)
        vmat = linalg.inv(vmat)
        vxmat = np.dot(vmat, xmat)
        xvxmat = np.dot(xmat.T, vxmat)
        xvxmat = linalg.inv(xvxmat)
        vxxvxxv_mat = reduce(np.dot, [vxmat, xvxmat, vxmat.T])
        pmat = vmat - vxxvxxv_mat
        pymat = np.dot(pmat, y)
        # Updated variances
        trace_term = np.trace(h_mat) * var_com[-1] - np.sum(h_mat*pmat) * var_com[-1] * var_com[-1]
        y_zu = y - np.dot(zgzmat, pymat)
        var_com_new[-1] = trace_term + np.sum(reduce(np.dot, [y_zu.T, h_mat, y_zu]))
        var_com_new[-1] /= n - xmat.shape[1]
        for k in range(num_var - 1):
            var_com_new[k] = -np.sum(zgzmat_lst[k] * pmat) + np.sum(reduce(np.dot, [pymat.T, zgzmat_lst[k], pymat]))
            var_com_new[k] = var_com[k] + var_com[k] * var_com[k] / gmat_lst[k].shape[0] * var_com_new[k]
        delta = var_com_new - var_com
        cc_par_val = np.sum(delta * delta) / np.sum(var_com_new * var_com_new)
        cc_par_val = np.sqrt(cc_par_val)
        var_com = np.array(var_com_new)
        logging.info('Norm of update vector: ' + str(cc_par_val))
        logging.info('Updated variances: ' + ' '.join(list(np.array(var_com_new, dtype=str))))
        if cc_par_val < cc_par:
            break
    if cc_par_val < cc_par:
        logging.info('Variances converged.')
    else:
        logging.info('Variances not converged.')
    return var_com



import numpy as np
from pandas_plink import read_plink1_bin
from tqdm import tqdm
import logging


def read_plink(bed_file):
    snp_info = read_plink1_bin(bed_file + ".bed", bed_file + ".bim", bed_file + ".fam", verbose=False)
    return snp_info.values


def impute_geno(snp_mat):
    ind_na = np.where(np.isnan(snp_mat))
    col_na = set(ind_na[1])  # get col number where na exist
    for i in tqdm(col_na):
        snpi = snp_mat[:, i]
        code0 = np.sum(np.absolute(snpi - 0.0) < 1e-10)
        code1 = np.sum(np.absolute(snpi - 1.0) < 1e-10)
        code2 = np.sum(np.absolute(snpi - 2.0) < 1e-10)
        code_count = code0 + code1 + code2
        p_lst = [code0/code_count, code1/code_count, code2/code_count]
        icol_na = np.where(np.isnan(snpi))
        snpi[icol_na] = np.random.choice([0.0, 1.0, 2.0], len(icol_na[0]), p=p_lst)
        snp_mat[:, i] = snpi
    return snp_mat


def shuffle_bed(bed_file):
    """
    shuffle the genotypes of individuals snp-by-snp
    :param bed_file: the prefix for plink binary file
    :return: the shuffled plink binary file
    """
    try:
        from pysnptools.snpreader import Bed
    except Exception as e:
        print(e)
        return 0
    logging.INFO('Read the plink file')
    data = Bed(bed_file, count_A1=False).read()
    num_snp = data.val.shape[1]
    logging.INFO("Start shuffle the genotypes snp-by-snp")
    for i in tqdm(range(num_snp)):
        np.random.shuffle(data.val[:, i])
    logging.INFO('Write the shuffled plink file')
    Bed.write(bed_file + '_shuffle', data, count_A1=False)
    return 1



from gmat.remma.remma_epiAA.remma_epiAA import remma_epiAA, remma_epiAA_parallel
from gmat.remma.remma_epiAA.remma_epiAA_pair import remma_epiAA_pair
from gmat.remma.remma_epiAA.remma_epiAA_eff import remma_epiAA_eff, remma_epiAA_eff_parallel
from gmat.remma.remma_epiAA.remma_epiAA_approx import remma_epiAA_approx, remma_epiAA_approx_parallel
from gmat.remma.remma_epiAA.remma_epiAA_maf_approx import remma_epiAA_maf_approx, remma_epiAA_maf_approx_parallel


from gmat.remma.remma_epiAA.remma_epiAA import _remma_epiAA, _remma_epiAA_parallel
from gmat.remma.remma_epiAA.remma_epiAA_pair import _remma_epiAA_pair
from gmat.remma.remma_epiAA.remma_epiAA_eff import _remma_epiAA_eff, _remma_epiAA_eff_parallel



import numpy as np
from scipy import linalg
from scipy import sparse
from scipy.sparse import isspmatrix
from .common import *


def pre_fd_mat(cmati, kin, covi_mat, eff, eff_ind, e, cov_dim, zmat_con_lst, wmat, num_record, var_com):
    fd_mat = []
    for i in range(len(cov_dim)):
        for j in range(cov_dim[i]):
            for k in range(j + 1):
                var_fd = np.zeros((cov_dim[i], cov_dim[i]))
                var_fd[j, k] = var_fd[k, j] = 1.0
                fd1 = np.sum(np.multiply(var_fd, covi_mat[i])) * zmat_con_lst[i].shape[1] / cov_dim[i]
                if isspmatrix(kin[i]):
                    temp = np.dot(covi_mat[i], np.dot(var_fd, covi_mat[i]))
                    temp = sparse.kron(temp, kin[i])
                    fd2 = np.sum(temp.multiply(cmati[eff_ind[i + 1][0]:eff_ind[i + 1][-1], \
                                               eff_ind[i + 1][0]:eff_ind[i + 1][-1]]))
                    fd3 = np.sum(np.dot(eff[eff_ind[i + 1][0]:eff_ind[i + 1][-1], :].T, \
                                        temp.dot(eff[eff_ind[i + 1][0]:eff_ind[i + 1][-1], :])))
                else:
                    temp = np.dot(covi_mat[i], np.dot(var_fd, covi_mat[i]))
                    temp = linalg.kron(temp, kin[i])
                    fd2 = np.sum(np.multiply(temp, cmati[eff_ind[i + 1][0]:eff_ind[i + 1][-1], \
                                                   eff_ind[i + 1][0]:eff_ind[i + 1][-1]]))
                    fd3 = np.sum(np.dot(eff[eff_ind[i + 1][0]:eff_ind[i + 1][-1], :].T, \
                                        np.dot(temp, eff[eff_ind[i + 1][0]:eff_ind[i + 1][-1], :])))
                fd_mat.append(-0.5 * (fd1 - fd2 - fd3))
    
    temp = -0.5 * (num_record / var_com[-1] - np.sum((wmat.T.dot(wmat)).multiply(cmati)) \
                   / (var_com[-1] * var_com[-1]) - np.sum(np.dot(e.T, e) / (var_com[-1] * var_com[-1])))
    fd_mat.append(temp)
    # print "FD matrix:", fd_mat
    fd_mat = np.array(fd_mat)
    return fd_mat


def pre_fd_mat_x(cmati, kin, covi_mat, eff, eff_ind, e, cov_dim, zmat_con_lst, wmat, num_record, var_com):
    fd_mat = []
    for i in range(len(cov_dim)):
        qnum = zmat_con_lst[i].shape[1] / cov_dim[i]
        tmat = np.array([0.0] * cov_dim[i] * cov_dim[i]).reshape(cov_dim[i], cov_dim[i])
        eff_lst = []
        if not isspmatrix(kin[i]):
            for j in range(cov_dim[i]):
                for k in range(j + 1):
                    tmat[j, k] = np.sum(np.multiply(kin[i], cmati[eff_ind[i + 1][j]:eff_ind[i + 1][j + 1],
                                                            eff_ind[i + 1][k]:eff_ind[i + 1][k + 1]]))
                    tmat[k, j] = tmat[j, k]
                eff_lst.append(eff[eff_ind[i + 1][j]:eff_ind[i + 1][j + 1], :])
            eff_mat = np.dot(np.concatenate(eff_lst, axis=1), covi_mat[i])
            temp = qnum * covi_mat[i] - np.dot(np.dot(covi_mat[i], tmat), covi_mat[i]) \
                   - np.dot(np.dot(eff_mat.T, kin[i]), eff_mat)
        else:
            for j in range(cov_dim[i]):
                for k in range(j + 1):
                    tmat[j, k] = np.sum(kin[i].multiply(cmati[eff_ind[i + 1][j]:eff_ind[i + 1][j + 1],
                                                        eff_ind[i + 1][k]:eff_ind[i + 1][k + 1]]))
                    tmat[k, j] = tmat[j, k]
                eff_lst.append(eff[eff_ind[i + 1][j]:eff_ind[i + 1][j + 1], :])
            eff_mat = np.dot(np.concatenate(eff_lst, axis=1), covi_mat[i])
            temp = qnum * covi_mat[i] - np.dot(np.dot(covi_mat[i], tmat), covi_mat[i]) \
                   - np.dot(eff_mat.T, kin[i].dot(eff_mat))
        
        temp[np.tril_indices(cov_dim[i], k=-1)] = 2.0 * temp[np.tril_indices(cov_dim[i], k=-1)]
        fd_mat.extend(list(-0.5 * temp[np.tril_indices(cov_dim[i], k=0)]))
    
    temp = -0.5 * (num_record / var_com[-1] - np.sum((wmat.T.dot(wmat)).multiply(cmati)) \
                   / (var_com[-1] * var_com[-1]) - np.sum(np.dot(e.T, e) / (var_com[-1] * var_com[-1])))
    
    fd_mat.append(temp)
    # print "FD matrix:", fd_mat
    fd_mat = np.array(fd_mat)
    return fd_mat


def pre_fd_sparse_mat(cmati, kin, covi_mat, eff, eff_ind, e, cov_dim, zmat_con_lst, wmat, num_record, var_com):
    fd_mat = []
    for i in range(len(cov_dim)):
        qnum = zmat_con_lst[i].shape[1] / cov_dim[i]
        tmat = np.array([0.0] * cov_dim[i] * cov_dim[i]).reshape(cov_dim[i], cov_dim[i])
        eff_lst = []
        for j in range(cov_dim[i]):
            for k in range(j + 1):
                tmat[j, k] = np.sum(kin[i].multiply(cmati[eff_ind[i + 1][j]:eff_ind[i + 1][j + 1],
                                                       eff_ind[i + 1][k]:eff_ind[i + 1][k + 1]]))
                tmat[k, j] = tmat[j, k]
            eff_lst.append(eff[eff_ind[i + 1][j]:eff_ind[i + 1][j + 1], :])
        eff_mat = np.dot(np.concatenate(eff_lst, axis=1), covi_mat[i])
        temp = qnum * covi_mat[i] - np.dot(np.dot(covi_mat[i], tmat), covi_mat[i]) \
                   -np.dot(eff_mat.T, kin[i].dot(eff_mat))
        temp[np.tril_indices(cov_dim[i], k=-1)] = 2.0 * temp[np.tril_indices(cov_dim[i], k=-1)]
        fd_mat.extend(list(-0.5 * temp[np.tril_indices(cov_dim[i], k=0)]))
    
    temp = -0.5 * (num_record / var_com[-1] - np.sum((wmat.T.dot(wmat)).multiply(cmati)) / (var_com[-1] * var_com[
        -1]) - np.sum(np.dot(e.T, e) / (var_com[-1] * var_com[-1])))
    
    fd_mat.append(temp)
    fd_mat = np.array(fd_mat)
    return fd_mat


def pre_ai_mat(cmati, covi_mat, eff, eff_ind, e, cov_dim, zmat_con_lst, wmat, var_com):
    ###working variable
    wv = []
    for i in range(len(cov_dim)):
        dial = sparse.eye(zmat_con_lst[i].shape[1] / cov_dim[i], dtype=np.float64)
        for j in range(cov_dim[i]):
            for k in range(j + 1):
                var_fd = np.zeros((cov_dim[i], cov_dim[i]))  # Notice
                var_fd[j, k] = var_fd[k, j] = 1.0
                temp = sparse.kron(np.dot(var_fd, covi_mat[i]), dial)
                temp = temp.dot(eff[eff_ind[i + 1][0]:eff_ind[i + 1][-1], :])
                temp = zmat_con_lst[i].dot(temp)
                wv.append(temp)
    wv.append(e / var_com[-1])
    qmat = np.concatenate(wv, axis=1)
    # print "Working variables: ", Q[:,:5]
    
    ###AI matrix
    qrq = np.divide(np.dot(qmat.T, qmat), var_com[-1])
    left = np.divide(wmat.T.dot(qmat), var_com[-1])
    eff_qmat = np.dot(cmati, left)
    ai_mat = np.subtract(qrq, np.dot(left.T, eff_qmat))
    ai_mat = 0.5 * ai_mat
    try:
        linalg.cholesky(ai_mat)
    except:
        ai_mat = False
    return ai_mat


def pre_ai_sparse_mat(cmati, covi_mat, eff, eff_ind, e, cov_dim, zmat_con_lst, wmat, var_com):
    ###working variable
    wv = []
    for i in range(len(cov_dim)):
        dial = sparse.eye(zmat_con_lst[i].shape[1] / cov_dim[i], dtype=np.float64)
        for j in range(cov_dim[i]):
            for k in range(j + 1):
                var_fd = np.zeros((cov_dim[i], cov_dim[i]))  # Notice
                var_fd[j, k] = var_fd[k, j] = 1.0
                temp = sparse.kron(np.dot(var_fd, covi_mat[i]), dial)
                temp = temp.dot(eff[eff_ind[i + 1][0]:eff_ind[i + 1][-1], :])
                temp = zmat_con_lst[i].dot(temp)
                wv.append(temp)
    wv.append(e / var_com[-1])
    qmat = np.concatenate(wv, axis=1)
    # print "Working variables: ", Q[:,:5]
    
    ###AI matrix
    qrq = np.divide(np.dot(qmat.T, qmat), var_com[-1])
    left = np.divide(wmat.T.dot(qmat), var_com[-1])
    eff_qmat = cmati.dot(left)
    ai_mat = np.subtract(qrq, np.dot(left.T, eff_qmat))
    ai_mat = 0.5 * ai_mat
    try:
        linalg.cholesky(ai_mat)
    except:
        ai_mat = False
    return ai_mat


def pre_em_mat(cov_dim, zmat_con_lst, num_record, var_com):
    num_var_com = len(var_com)
    em_mat = np.array([0.0] * num_var_com * num_var_com).reshape(num_var_com, num_var_com)
    b = 0
    cov_mat = []
    for i in range(len(cov_dim)):
        ind = np.tril_indices(cov_dim[i])
        temp = np.array([0.0] * len(ind[0]) * len(ind[0])).reshape(len(ind[0]), len(ind[0]))
        q = zmat_con_lst[i].shape[1] / cov_dim[i]
        a = b
        b = b + len(ind[0])
        covar = np.zeros((cov_dim[i], cov_dim[i]))
        covar[ind] = var_com[a:b]
        covar = np.add(covar, np.tril(covar, -1).T)
        cov_mat.append(covar)
        for j in range(len(ind[0])):
            for k in range(j + 1):
                temp[j, k] = (cov_mat[i][ind[0][j], ind[0][k]] * cov_mat[i][ind[1][j], ind[1][k]] + \
                              cov_mat[i][ind[0][j], ind[1][k]] * cov_mat[i][ind[1][j], ind[0][k]]) / (2.0 * q)
        em_mat[a:b, a:b] = temp.copy()
    em_mat[-1, -1] = (var_com[-1] * var_com[-1]) / num_record
    em_mat = 2.0 * em_mat
    
    em_mat += np.tril(em_mat, k=-1).T
    return linalg.inv(em_mat)


def pre_fdai_mat_eigen(y, xmat, leg_tp, kin_eigen_val, cov_add, cov_per, cov_res, var_com, var_ind, kin_dct):
    vinv = np.multiply(kin_eigen_val.reshape(len(kin_eigen_val), 1, 1), tri_matT(leg_tp, cov_add)) + \
           tri_matT(leg_tp, cov_per) + np.diag([np.sum(cov_res)] * leg_tp.shape[0])
    vinv = np.linalg.inv(vinv)
    
    xv = np.matmul(xmat.transpose(0, 2, 1), vinv)
    xvx = np.matmul(xv, xmat)
    xvx = reduce(np.add, xvx)
    xvx = np.linalg.inv(xvx)
    
    # xvy
    yv = np.matmul(y.transpose(0, 2, 1), vinv)
    yvx = np.matmul(yv, xmat)
    yvx = reduce(np.add, yvx)
    
    fd_mat = np.array([0.0] * len(var_com))
    ai_mat = np.array([0.0] * len(var_com) * len(var_com)).reshape(len(var_com), len(var_com))
    
    for m in range(var_ind.shape[0]):
        
        onei = np.array([0.0] * leg_tp.shape[1]).reshape(leg_tp.shape[1], 1)
        onej = onei.copy()
        onei[var_ind[m, 1], 0] = 1.0
        onej[var_ind[m, 2], 0] = 1.0
        one = np.dot(onei, onej.T)
        one = one + one.T - np.diag(np.diag(one))
        tit = reduce(np.dot, [leg_tp, one, leg_tp.T])
        if m == var_ind.shape[0] - 1:
            tit = np.eye(leg_tp.shape[0])
        
        # tr(vk)
        trvk = np.matmul(vinv, tit)
        trvk = np.multiply(kin_dct[var_ind[m, 0]], trvk)
        trvk = np.trace(reduce(np.add, trvk))
        
        # tr(pk)
        xvkvx = reduce(np.matmul, [xv, tit, xv.transpose(0, 2, 1)])
        xvkvx = np.multiply(kin_dct[var_ind[m, 0]], xvkvx)
        xvkvx = reduce(np.add, xvkvx)
        trpk = np.sum(np.multiply(xvx, xvkvx))
        # ypkpy
        yvkvy = reduce(np.matmul, [yv, tit, yv.transpose(0, 2, 1)])
        yvkvy = np.multiply(kin_dct[var_ind[m, 0]], yvkvy)
        yvkvy = reduce(np.add, yvkvy)
        yvkvx = reduce(np.matmul, [yv, tit, xv.transpose(0, 2, 1)])
        yvkvx = np.multiply(kin_dct[var_ind[m, 0]], yvkvx)
        yvkvx = reduce(np.add, yvkvx)
        ypkpy2 = reduce(np.dot, [yvx, xvx, yvkvx.T])
        ypkpy3 = np.dot(yvx, xvx)
        ypkpy3 = reduce(np.dot, [ypkpy3, xvkvx, ypkpy3.T])
        ypkpy = np.sum(yvkvy) - 2 * np.sum(ypkpy2) + np.sum(ypkpy3)
        fd_mat[m] = -0.5 * (trvk - trpk) + 0.5 * ypkpy
        
        for n in range(m + 1):
            onei2 = np.array([0.0] * leg_tp.shape[1]).reshape(leg_tp.shape[1], 1)
            onej2 = onei2.copy()
            onei2[var_ind[n, 1], 0] = 1.0
            onej2[var_ind[n, 2], 0] = 1.0
            one2 = np.dot(onei2, onej2.T)
            one2 = one2 + one2.T - np.diag(np.diag(one2))
            tit2 = reduce(np.dot, [leg_tp, one2, leg_tp.T])
            if n == var_ind.shape[0] - 1:
                tit2 = np.eye(leg_tp.shape[0])
            yvkvkvy = reduce(np.matmul, [yv, tit, vinv, tit2, yv.transpose(0, 2, 1)])
            yvkvkvy = reduce(np.multiply, [kin_dct[var_ind[m, 0]], kin_dct[var_ind[n, 0]], yvkvkvy])
            val1 = reduce(np.add, yvkvkvy)
            
            yvkvkvx = reduce(np.matmul, [yv, tit, vinv, tit2, xv.transpose(0, 2, 1)])
            yvkvkvx = reduce(np.multiply, [kin_dct[var_ind[m, 0]], kin_dct[var_ind[n, 0]], yvkvkvx])
            yvkvkvx = reduce(np.add, yvkvkvx)
            val2 = reduce(np.dot, [yvkvkvx, xvx, yvx.T])
            
            yvkvx2 = reduce(np.matmul, [yv, tit2, xv.transpose(0, 2, 1)])
            yvkvx2 = np.multiply(kin_dct[var_ind[n, 0]], yvkvx2)
            yvkvx2 = reduce(np.add, yvkvx2)
            val3 = reduce(np.dot, [yvkvx, xvx, yvkvx2.T])
            
            xvkvx2 = reduce(np.matmul, [xv, tit2, xv.transpose(0, 2, 1)])
            xvkvx2 = np.multiply(kin_dct[var_ind[n, 0]], xvkvx2)
            xvkvx2 = reduce(np.add, xvkvx2)
            val4 = reduce(np.dot, [yvkvx, xvx, xvkvx2, xvx, yvx.T])
            
            yvkvkvx2 = reduce(np.matmul, [yv, tit2, vinv, tit, xv.transpose(0, 2, 1)])
            yvkvkvx2 = reduce(np.multiply, [kin_dct[var_ind[m, 0]], kin_dct[var_ind[n, 0]], yvkvkvx2])
            yvkvkvx2 = reduce(np.add, yvkvkvx2)
            val5 = reduce(np.dot, [yvx, xvx, yvkvkvx2.T])
            
            xvkvkvx2 = reduce(np.matmul, [xv, tit, vinv, tit2, xv.transpose(0, 2, 1)])
            xvkvkvx2 = reduce(np.multiply, [kin_dct[var_ind[m, 0]], kin_dct[var_ind[n, 0]], xvkvkvx2])
            xvkvkvx2 = reduce(np.add, xvkvkvx2)
            val6 = reduce(np.dot, [yvx, xvx, xvkvkvx2, xvx, yvx.T])
            
            val7 = reduce(np.dot, [yvx, xvx, xvkvx, xvx, yvkvx2.T])
            val8 = reduce(np.dot, [yvx, xvx, xvkvx, xvx, xvkvx2, xvx, yvx.T])
            temp = val1 - val2 - val3 + val4 - val5 + val6 + val7 - val8
            ai_mat[m, n] = 0.5 * np.sum(temp)
    ai_mat = ai_mat + np.tril(ai_mat, k=-1).T
    
    return fd_mat, ai_mat


def pre_em_mat_eigen(cov_dim, cov_add, cov_per, ran_df, var_com, num_record):
    num_var_com = len(var_com)
    em_mat = np.array([0.0] * num_var_com * num_var_com).reshape(num_var_com, num_var_com)
    
    ind = np.tril_indices(cov_dim)
    temp = np.array([0.0] * len(ind[0]) * len(ind[0])).reshape(len(ind[0]), len(ind[0]))
    for j in range(len(ind[0])):
        for k in range(j + 1):
            temp[k, j] = temp[j, k] = (cov_add[ind[0][j], ind[0][k]] * cov_add[ind[1][j], ind[1][k]] + \
                                       cov_add[ind[0][j], ind[1][k]] * cov_add[ind[1][j], ind[0][k]]) / (2.0 * ran_df)
    temp = linalg.inv(temp)
    em_mat[0:len(ind[0]), 0:len(ind[0])] = temp.copy()
    
    temp = np.array([0.0] * len(ind[0]) * len(ind[0])).reshape(len(ind[0]), len(ind[0]))
    for j in range(len(ind[0])):
        for k in range(j + 1):
            temp[k, j] = temp[j, k] = (cov_per[ind[0][j], ind[0][k]] * cov_per[ind[1][j], ind[1][k]] + \
                                       cov_per[ind[0][j], ind[1][k]] * cov_per[ind[1][j], ind[0][k]]) / (2.0 * ran_df)
    temp = linalg.inv(temp)
    em_mat[len(ind[0]):2 * len(ind[0]), len(ind[0]):2 * len(ind[0])] = temp.copy()
    
    em_mat[-1, -1] = num_record / (var_com[-1] * var_com[-1])
    
    em_mat = em_mat / 2.0
    return em_mat


def pre_fdai_mat_eigen_x(y, xmat, leg_tp, kin_eigen_val, cov_add, cov_per, cov_res, var_com, var_ind, kin_dct):
    vinv = np.multiply(kin_eigen_val.reshape(len(kin_eigen_val), 1, 1), tri_matT(leg_tp, cov_add)) + \
           tri_matT(leg_tp, cov_per) + np.diag([np.sum(cov_res)] * leg_tp.shape[0])
    vinv = np.linalg.inv(vinv)
    
    xvx = reduce(np.matmul, [xmat.transpose(0, 2, 1), vinv, xmat])
    xvx = reduce(np.add, xvx)
    xvx = np.linalg.inv(xvx)
    
    # xvy
    yv = np.matmul(y.transpose(0, 2, 1), vinv)
    yvx = np.matmul(yv, xmat)
    yvx = reduce(np.add, yvx)
    
    yvvx = reduce(np.matmul, [yv, vinv, xmat])
    yvvx = reduce(np.add, yvvx)
    
    xvvx = reduce(np.matmul, [xmat.transpose(0, 2, 1), vinv, vinv, xmat])
    xvvx = reduce(np.add, xvvx)
    
    fd_mat = np.array([0.0] * len(var_com))
    ai_mat = np.array([0.0] * len(var_com) * len(var_com)).reshape(len(var_com), len(var_com))
    
    for m in range(var_ind.shape[0] - 1):
        onei = np.array([0.0] * leg_tp.shape[1]).reshape(leg_tp.shape[1], 1)
        onej = onei.copy()
        onei[var_ind[m, 1], 0] = 1.0
        onej[var_ind[m, 2], 0] = 1.0
        one = np.dot(onei, onej.T)
        one = one + one.T - np.diag(np.diag(one))
        tit = reduce(np.dot, [leg_tp, one, leg_tp.T])
        
        # tr(vk)
        trvk = reduce(np.matmul, [onej.T, leg_tp.T, vinv, leg_tp, onei])
        trvk = np.multiply(kin_dct[var_ind[m, 0]], trvk)
        trvk = np.sum(reduce(np.add, trvk))
        
        if var_ind[m, 1] != var_ind[m, 2]:
            trvk = 2 * trvk
        
        # tr(pk)
        temp2 = reduce(np.matmul, [onej.T, leg_tp.T, vinv, xmat])
        temp1 = reduce(np.matmul, [onei.T, leg_tp.T, vinv, xmat])
        xvkvx = np.matmul(temp1.transpose(0, 2, 1), temp2)
        xvkvx = np.multiply(kin_dct[var_ind[m, 0]], xvkvx)
        xvkvx = reduce(np.add, xvkvx)
        if var_ind[m, 1] != var_ind[m, 2]:
            xvkvx = xvkvx + xvkvx.T
        trpk = np.sum(np.multiply(xvx, xvkvx))
        
        # ypkpy
        yvkvy = reduce(np.matmul, [yv, tit, yv.transpose(0, 2, 1)])
        yvkvy = np.multiply(kin_dct[var_ind[m, 0]], yvkvy)
        yvkvy = reduce(np.add, yvkvy)
        yvkvx = reduce(np.matmul, [yv, tit, vinv, xmat])
        yvkvx = np.multiply(kin_dct[var_ind[m, 0]], yvkvx)
        yvkvx = reduce(np.add, yvkvx)
        ypkpy2 = reduce(np.dot, [yvx, xvx, yvkvx.T])
        ypkpy3 = np.dot(yvx, xvx)
        ypkpy3 = reduce(np.dot, [ypkpy3, xvkvx, ypkpy3.T])
        ypkpy = np.sum(yvkvy) - 2 * np.sum(ypkpy2) + np.sum(ypkpy3)
        fd_mat[m] = -0.5 * (trvk - trpk) + 0.5 * ypkpy
        
        for n in range(m + 1):
            onei2 = np.array([0.0] * leg_tp.shape[1]).reshape(leg_tp.shape[1], 1)
            onej2 = onei2.copy()
            onei2[var_ind[n, 1], 0] = 1.0
            onej2[var_ind[n, 2], 0] = 1.0
            one2 = np.dot(onei2, onej2.T)
            one2 = one2 + one2.T - np.diag(np.diag(one2))
            tit2 = reduce(np.dot, [leg_tp, one2, leg_tp.T])
            yvkvkvy = reduce(np.matmul, [yv, tit, vinv, tit2, yv.transpose(0, 2, 1)])
            yvkvkvy = reduce(np.multiply, [kin_dct[var_ind[m, 0]], kin_dct[var_ind[n, 0]], yvkvkvy])
            val1 = reduce(np.add, yvkvkvy)
            
            yvkvkvx = reduce(np.matmul, [yv, tit, vinv, tit2, vinv, xmat])
            yvkvkvx = reduce(np.multiply, [kin_dct[var_ind[m, 0]], kin_dct[var_ind[n, 0]], yvkvkvx])
            yvkvkvx = reduce(np.add, yvkvkvx)
            val2 = reduce(np.dot, [yvkvkvx, xvx, yvx.T])
            
            yvkvx2 = reduce(np.matmul, [yv, tit2, vinv, xmat])
            yvkvx2 = np.multiply(kin_dct[var_ind[n, 0]], yvkvx2)
            yvkvx2 = reduce(np.add, yvkvx2)
            val3 = reduce(np.dot, [yvkvx, xvx, yvkvx2.T])
            
            temp2 = reduce(np.matmul, [onej2.T, leg_tp.T, vinv, xmat])
            temp1 = reduce(np.matmul, [onei2.T, leg_tp.T, vinv, xmat])
            xvkvx2 = np.matmul(temp1.transpose(0, 2, 1), temp2)
            xvkvx2 = np.multiply(kin_dct[var_ind[n, 0]], xvkvx2)
            xvkvx2 = reduce(np.add, xvkvx2)
            if var_ind[n, 1] != var_ind[n, 2]:
                xvkvx2 = xvkvx2 + xvkvx2.T
            val4 = reduce(np.dot, [yvkvx, xvx, xvkvx2, xvx, yvx.T])
            
            yvkvkvx2 = reduce(np.matmul, [yv, tit2, vinv, tit, vinv, xmat])
            yvkvkvx2 = reduce(np.multiply, [kin_dct[var_ind[m, 0]], kin_dct[var_ind[n, 0]], yvkvkvx2])
            yvkvkvx2 = reduce(np.add, yvkvkvx2)
            val5 = reduce(np.dot, [yvx, xvx, yvkvkvx2.T])
            
            xvkvkvx2 = cal_xvkvkvx(xmat, vinv, onei, onej, onei2, onej2, leg_tp, kin_dct, var_ind, m, n)
            if var_ind[n, 1] != var_ind[n, 2] and var_ind[m, 1] != var_ind[m, 2]:
                xvkvkvx2 = xvkvkvx2 + \
                           cal_xvkvkvx(xmat, vinv, onej, onei, onei2, onej2, leg_tp, kin_dct, var_ind, m, n) \
                           + cal_xvkvkvx(xmat, vinv, onei, onej, onej2, onei2, leg_tp, kin_dct, var_ind, m, n) \
                           + cal_xvkvkvx(xmat, vinv, onej, onei, onej2, onei2, leg_tp, kin_dct, var_ind, m, n)
            elif var_ind[m, 1] != var_ind[m, 2]:
                xvkvkvx2 = xvkvkvx2 + cal_xvkvkvx(xmat, vinv, onej, onei, onei2, onej2, leg_tp, kin_dct, var_ind, m, n)
            elif var_ind[n, 1] != var_ind[n, 2]:
                xvkvkvx2 = xvkvkvx2 + cal_xvkvkvx(xmat, vinv, onei, onej, onej2, onei2, leg_tp, kin_dct, var_ind, m, n)
            val6 = reduce(np.dot, [yvx, xvx, xvkvkvx2, xvx, yvx.T])
            
            val7 = reduce(np.dot, [yvx, xvx, xvkvx, xvx, yvkvx2.T])
            val8 = reduce(np.dot, [yvx, xvx, xvkvx, xvx, xvkvx2, xvx, yvx.T])
            temp = val1 - val2 - val3 + val4 - val5 + val6 + val7 - val8
            ai_mat[m, n] = 0.5 * np.sum(temp)
        
        yvkvkvy = reduce(np.matmul, [yv, tit, vinv, yv.transpose(0, 2, 1)])
        yvkvkvy = reduce(np.multiply, [kin_dct[var_ind[m, 0]], yvkvkvy])
        val1 = reduce(np.add, yvkvkvy)
        
        yvkvkvx = reduce(np.matmul, [yv, tit, vinv, vinv, xmat])
        yvkvkvx = reduce(np.multiply, [kin_dct[var_ind[m, 0]], yvkvkvx])
        yvkvkvx = reduce(np.add, yvkvkvx)
        val2 = reduce(np.dot, [yvkvkvx, xvx, yvx.T])
        
        val3 = reduce(np.dot, [yvkvx, xvx, yvvx.T])
        
        val4 = reduce(np.dot, [yvkvx, xvx, xvvx, xvx, yvx.T])
        
        yvkvkvx2 = reduce(np.matmul, [yv, vinv, tit, vinv, xmat])
        yvkvkvx2 = reduce(np.multiply, [kin_dct[var_ind[m, 0]], yvkvkvx2])
        yvkvkvx2 = reduce(np.add, yvkvkvx2)
        val5 = reduce(np.dot, [yvx, xvx, yvkvkvx2.T])
        
        xvkvkvx2 = cal_xvkvkvx_res(xmat, vinv, onei, onej, leg_tp, kin_dct, var_ind, m)
        if var_ind[m, 1] != var_ind[m, 2]:
            xvkvkvx2 = xvkvkvx2 + cal_xvkvkvx_res(xmat, vinv, onej, onei, leg_tp, kin_dct, var_ind, m)
        val6 = reduce(np.dot, [yvx, xvx, xvkvkvx2, xvx, yvx.T])
        
        val7 = reduce(np.dot, [yvx, xvx, xvkvx, xvx, yvvx.T])
        val8 = reduce(np.dot, [yvx, xvx, xvkvx, xvx, xvvx, xvx, yvx.T])
        temp = val1 - val2 - val3 + val4 - val5 + val6 + val7 - val8
        ai_mat[-1, m] = 0.5 * np.sum(temp)
    
    # tr(vk)
    trvk = np.trace(reduce(np.add, vinv))
    # tr(pk)
    trpk = np.sum(np.multiply(xvx, xvvx))
    # ypkpy
    yvvy = reduce(np.matmul, [yv, yv.transpose(0, 2, 1)])
    yvvy = reduce(np.add, yvvy)
    ypkpy2 = reduce(np.dot, [yvx, xvx, yvvx.T])
    ypkpy3 = np.dot(yvx, xvx)
    ypkpy3 = reduce(np.dot, [ypkpy3, xvvx, ypkpy3.T])
    ypkpy = np.sum(yvvy) - 2 * np.sum(ypkpy2) + np.sum(ypkpy3)
    fd_mat[-1] = -0.5 * (trvk - trpk) + 0.5 * ypkpy
    
    # ai_mat
    yvvvy = reduce(np.matmul, [yv, vinv, yv.transpose(0, 2, 1)])
    val1 = reduce(np.add, yvvvy)
    yvvvx = reduce(np.matmul, [yv, vinv, vinv, xmat])
    yvvvx = reduce(np.add, yvvvx)
    val2 = reduce(np.dot, [yvvvx, xvx, yvx.T])
    val3 = reduce(np.dot, [yvvx, xvx, yvvx.T])
    val4 = reduce(np.dot, [yvvx, xvx, xvvx, xvx, yvx.T])
    val5 = reduce(np.dot, [yvx, xvx, yvvvx.T])
    xvvvx = reduce(np.matmul, [xmat.transpose(0, 2, 1), vinv, vinv, vinv, xmat])
    xvvvx = reduce(np.add, xvvvx)
    val6 = reduce(np.dot, [yvx, xvx, xvvvx, xvx, yvx.T])
    val7 = reduce(np.dot, [yvx, xvx, xvvx, xvx, yvvx.T])
    val8 = reduce(np.dot, [yvx, xvx, xvvx, xvx, xvvx, xvx, yvx.T])
    temp = val1 - val2 - val3 + val4 - val5 + val6 + val7 - val8
    ai_mat[-1, -1] = 0.5 * np.sum(temp)
    
    ai_mat = ai_mat + np.tril(ai_mat, k=-1).T
    
    return fd_mat, ai_mat


def pre_fdai_mat_eigen_glm(y, xmat, leg_tp, kin_eigen_val, cov_add, cov_per, cov_res, var_com, var_ind, kin_dct):
    vinv = np.multiply(kin_eigen_val.reshape(len(kin_eigen_val), 1, 1), tri_matT(leg_tp, cov_add)) + \
           tri_matT(leg_tp, cov_per) + np.diag([np.sum(cov_res)] * leg_tp.shape[0])
    vinv = np.linalg.inv(vinv)
    
    xvx = reduce(np.matmul, [xmat.transpose(0, 2, 1), vinv, xmat])
    xvx = reduce(np.add, xvx)
    xvx = np.linalg.inv(xvx)
    
    # xvy
    xvy = np.matmul(xmat.transpose(0, 2, 1), np.matmul(vinv, y))
    xvy = reduce(np.add, xvy)
    
    # py
    y_xb = y - np.matmul(xmat, np.dot(xvx, xvy))
    py = np.matmul(vinv, y_xb)
    
    fd_mat = np.array([0.0] * len(var_com))
    ai_mat = np.array([0.0] * len(var_com) * len(var_com)).reshape(len(var_com), len(var_com))
    ai_kpy = []
    ai_pkpy = []
    for m in range(var_ind.shape[0] - 1):
        
        onei = np.array([0.0] * leg_tp.shape[1]).reshape(leg_tp.shape[1], 1)
        onej = onei.copy()
        onei[var_ind[m, 1], 0] = 1.0
        onej[var_ind[m, 2], 0] = 1.0
        one = np.dot(onei, onej.T)
        one = one + one.T - np.diag(np.diag(one))
        tit = reduce(np.dot, [leg_tp, one, leg_tp.T])
        
        # tr(vk)
        trvk = reduce(np.matmul, [onej.T, leg_tp.T, vinv, leg_tp, onei])
        trvk = np.multiply(kin_dct[var_ind[m, 0]], trvk)
        trvk = np.sum(reduce(np.add, trvk))
        
        if var_ind[m, 1] != var_ind[m, 2]:
            trvk = 2 * trvk
        
        # tr(pk)
        temp2 = reduce(np.matmul, [onej.T, leg_tp.T, vinv, xmat])
        temp1 = reduce(np.matmul, [onei.T, leg_tp.T, vinv, xmat])
        xvkvx = np.matmul(temp1.transpose(0, 2, 1), temp2)
        xvkvx = np.multiply(kin_dct[var_ind[m, 0]], xvkvx)
        xvkvx = reduce(np.add, xvkvx)
        if var_ind[m, 1] != var_ind[m, 2]:
            xvkvx = xvkvx + xvkvx.T
        trpk = np.sum(np.multiply(xvx, xvkvx))
        
        kpy = np.matmul(tit, py)
        kpy = np.multiply(kin_dct[var_ind[m, 0]], kpy)
        ai_kpy.append(kpy)
        
        ypkpy = np.sum(np.matmul(py.transpose(0, 2, 1), kpy))
        fd_mat[m] = -0.5 * (trvk - trpk) + 0.5 * ypkpy
        
        xvkpy = np.matmul(xmat.transpose(0, 2, 1), np.matmul(vinv, kpy))
        xvkpy = reduce(np.add, xvkpy)
        
        kpy_xb = kpy - np.matmul(xmat, np.dot(xvx, xvkpy))
        pkpy = np.matmul(vinv, kpy_xb)
        ai_pkpy.append(pkpy)
    
    trvk = np.sum(map(np.trace, vinv))
    # tr(pk)
    xvvx = reduce(np.matmul, [xmat.transpose(0, 2, 1), vinv, vinv, xmat])
    xvvx = reduce(np.add, xvvx)
    trpk = np.sum(np.multiply(xvx, xvvx))
    ai_kpy.append(py)
    
    ypkpy = np.sum(np.matmul(py.transpose(0, 2, 1), py))
    fd_mat[-1] = -0.5 * (trvk - trpk) + 0.5 * ypkpy
    
    xvkpy = np.matmul(xmat.transpose(0, 2, 1), np.matmul(vinv, py))
    xvkpy = reduce(np.add, xvkpy)
    
    kpy_xb = py - np.matmul(xmat, np.dot(xvx, xvkpy))
    pkpy = np.matmul(vinv, kpy_xb)
    ai_pkpy.append(pkpy)
    
    for i in range(len(ai_kpy)):
        for j in range(i + 1):
            ai_mat[i, j] = np.sum(np.matmul(ai_kpy[i].transpose(0, 2, 1), ai_pkpy[j]))
    
    ai_mat = ai_mat + np.tril(ai_mat, k=-1).T
    ai_mat = 0.5 * ai_mat
    
    return fd_mat, ai_mat
